""" This is the actual code that wraps TOUGH2 model run with pre- and post-
processing, so all PEST's model/batch file call this.

Use with optiona flags:
    python pest_model.py [--svda] [--obsreref]

Run with "--svda" flag with additional command of parcalc for SVDassist runs.
Flag "--obsreref" is for observation re-referencing which reset save to incon
and overwrites the master's incon, which will be used by all subsequent model
runs.  Flag "--test-update" will cause each slave to save a unique
real_model.save file back to master directory.  This allows "--obsreref" to
later select the best save/incon to start with.

To generate/overwrite/fix the model/batch files, use:
    python make_batch_files.py
"""

import time
import sys
import glob
import os
from os import devnull, system, remove, sep, path
from shutil import copy2
from shutil import Error
from time import sleep

from numpy.testing import assert_approx_equal

from gopest.run_ns_pr import run_ns_pr
from gopest.par import generate_real_model
from gopest.obs import read_from_real_model

from gopest.common import config
from gopest.common import runtime

def get_master_dir():
    if config['mode'] != 'local':
        with open('_master_dir', 'r') as f:
            line = f.readlines()[0].strip()
            return line

def get_pest_dir():
    with open('_pest_dir', 'r') as f:
        line = f.readlines()[0].strip()
        return line

def get_t2():
    with open('_tough2', 'r') as f:
        line = f.readlines()[0].strip()
        return line

def get_slave_id():
    try:
        with open('_procid', 'r') as f:
            line = f.readlines()[0].strip()
            return line
    except:
        return '0'

def par_match(pf1, pf2):
    matched = False
    with open(pf1,'r') as a:
        with open(pf2,'r') as b:
            # once open successfully, assume equal, until something fails
            matched = True
            try:
                for aa,bb in zip(a,b):
                    ax, bx = float(aa.split(',')[0]), float(bb.split(',')[0])
                    assert_approx_equal(ax, bx, significant=7)
            except AssertionError:
                matched = False
    return matched

def main(obsreref, svda, testup, local, skiprun, useobf, sendbad, skippr, hdf5, waiwera):
    fgeo = runtime['filename']['geom']
    fsave = runtime['filename']['save']
    fincon = runtime['filename']['incon']
    fdato = runtime['filename']['dat_orig']
    fdats = runtime['filename']['dat_seq']
    flsts = runtime['filename']['lst_seq']

    print("----- Running " + " ".join(sys.argv[1:]))
    if local:
        master_dir = '.'
    else:
        master_dir = get_master_dir()
    print("--- Clean up pest_model.obf")
    if path.isfile('pest_model.obf'):
        remove('pest_model.obf')

    ### skips everything if use obf directly
    if useobf:
        raise Exception('Should use pest_hp file distribution 17/12/2022')
        if path.isfile('pest_model.obf.use'):
            copy2(master_dir + sep + 'pest_model.obf.use', 'pest_model.obf')
            print("Found pest_model.obf.use , skips everything.")
            return
        else:
            print("Error, cannot find existing pest_model.obf.use file")

    ### SVD-assist
    if svda:
        print("--- PARCALC")
        try:
            remove('pest_model.dat')
        except:
            print("pest_model.dat probably does not exist")
        PARCALC = path.join(get_pest_dir(), 'parcalc')
        system(PARCALC + ' > ' + devnull)

    ### goPESTpar
    if not skiprun:
        print("--- goPESTpar")
        generate_real_model(fdato, 'pest_model.dat', fdats[0])
        # sleep(30)  # just in case shared file system slow

    if obsreref:
        if path.isfile('pest_model.obf'):
            remove('pest_model.obf')
        # get matching incon, if exist
        for parf in glob.glob(master_dir + sep + 'pest_model.dat.*'):
            if par_match(parf, 'pest_model.dat'):
                matchname = path.splitext(parf)[1]
                print("--- found matched incon/pars %s from master dir, overwrite Master INCON" % matchname)
                copy2(master_dir + sep + fincon + matchname, master_dir + sep + fincon)
                copy2(master_dir + sep + 'pest_model.obf' + matchname, 'pest_model.obf')
                break
        # print("--- remove all pairs from labmda tests after searching")
        # for f in glob.glob(master_dir + sep + 'pest_model.obf.*'):
        #     remove(f)
        # for f in glob.glob(master_dir + sep + 'pest_model.dat.*'):
        #     remove(f)
        # for f in glob.glob(master_dir + sep + 'real_model.incon.*'):
        #     remove(f)
        if path.isfile('pest_model.obf'):
            print("--- use obf, skip actual model run")
            return
        else:
            print("--- could not find matching pars, obsreref continue with normal run")

    ### RUN TOUGH2 model
    if skiprun:
        print("--- skip actual TOUGH2 run")
    else:
        if not local:
            print("--- use master INCON")
            try:
                copy2(master_dir + sep + fincon, fincon)
            except Error as e:
                # OK if src and dst are the same file, simply skip.
                print(e)

        START_TIME = time.time()
        print("--- run_ns_pr.py")
        runok = run_ns_pr()
        if obsreref:
            if not local:
                print("--- reset Master INCON")
                copy2('real_model.incon', master_dir + sep + 'real_model.incon')
            else:
                print("--- .save file written as .incon")
        if sendbad:
            for f in glob.glob('bad_model_*'):
                copy2(f, master_dir + sep + 'bad_model_slave' + get_slave_id() + '_' + f)
        if not runok:
            print("--- run_ns_pr failed, skip goPESTobs, no obf, make sure lamforgive/derforgive is used.")
            return
        print('--- run_ns_pr complete after', (time.time() - START_TIME), 'seconds')

    ### goPESTobs
    # sleep(30)  # just in case shared file system slow
    print("--- goPESTobs")
    read_from_real_model(fgeo, fdats[-1], flsts[-1], 'pest_model.obf', waiwera=waiwera)

    if testup:
        print("--- store lambda test (save,obf,pars) pair:" + get_slave_id())
        copy2(fsave, master_dir + sep + fincon + '.' + get_slave_id())
        copy2('pest_model.dat', master_dir + sep + 'pest_model.dat.' + get_slave_id())
        copy2('pest_model.obf', master_dir + sep + 'pest_model.obf.' + get_slave_id())

def main_cli(argv=[]):
    obsreref = False
    svda = False
    testup = False
    skiprun = False
    useobf = False
    sendbad = True
    skippr = False
    waiwera = False
    hdf5 = False

    if config['simulator']['output-type'] == 'h5':
        hdf5 = True
    else:
        hdf5 = False
    
    if config['simulator']['input-type'] == 'waiwera':
        waiwera = True
        hdf5 = True
    else:
        waiwera = False

    local = config['mode'] == 'local'
    skiprun = config['model']['skip']
    skippr = config['model']['skip-pr']

    print('pest_model.py running at ', os.getcwd())
    
    if len(argv) > 1:
        if '--test-update' in argv[1:]:
            testup = True
        if '--obsreref' in argv[1:]:
            obsreref = True
        if '--svda' in argv[1:]:
            svda = True
        if '--use-obf' in argv[1:]:
            # requires existing pest_model.obf.use (PEST will remove
            # pest_model.obf, so use different name)
            useobf = True
        if '--local' in argv[1:]:
            local = True
    main(obsreref, svda, testup, local, skiprun, useobf, sendbad, skippr, hdf5, waiwera)



"""
Copyright 2013, 2014 University of Auckland.

This file is part of TIM (Tim Isn't Mulgraph).

    TIM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TIM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TIM.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
Wrap Waiwera's hdf5 output as t2listing
"""

import h5py
import numpy as np

from mulgrids import mulgrid
from t2listing import listingtable

import json
import time
import unittest
from pprint import pprint as pp

class wlisting(object):
    def __init__(self, filename=None, geo=None, fjson=None, size_check=True):
        """ Waiwera h5 output pretending to be t2listing

        If corresponding geo is supplied, wlisting can behave more like
        t2listing, which includes atmosphere blocks

        If Waiwera input json is provided, then .generation has .row_name using
        (t2 block name, source name) instead of (cell index, source index).
        """
        self._table = {}
        if isinstance(geo, str):
            self.geo = mulgrid(geo)
        else:
            self.geo = geo
        if self.geo is not None:
            if self.geo.block_order != 'dmplex':
                raise Exception("wlisting loading Waiwera output file requires a geometry with .block_order = 'dmplex'")
        if isinstance(fjson, str):
            with open(fjson, 'r') as f:
                self.wjson = json.load(f)
        else:
            self.wjson = json
        self._h5 = h5py.File(filename, 'r')
        self.filename = filename
        self.simulator = 'waiwera'
        self.size_check = size_check # raise Exception if number of block does not match geo
        self.setup()

    def setup(self):
        self.cell_idx = self._h5['cell_index'][:,0]
        self.fulltimes = self._h5['time'][:,0]
        self.num_fulltimes = len(self.fulltimes)
        self._index = 0
        ### checks:
        nh5 = len(self.cell_idx)
        if self.geo is not None:
            # print('wlisting.element: uses mulgrid block name (str) as key.')
            nb = self.geo.num_blocks - self.geo.num_atmosphere_blocks
            if self.size_check and nh5 != nb:
                msg = 'HDF5 result %s has %i cells different from geometry %s (%i excl. atmosphere blocks)' % (
                    self.filename, nh5, self.geo.filename, nb)
                raise Exception(msg)
            if nh5 < nb:
                msg = 'HDF5 result %s has %i cells less than from geometry %s (%i excl. atmosphere blocks)' % (
                    self.filename, nh5, self.geo.filename, nb)
                raise Exception(msg)
            # blocks is seen by TIM/geo, w_blocks is used by waiwera-h5/json
            blocks = self.geo.block_name_list
            w_blocks = self.geo.block_name_list[self.geo.num_atmosphere_blocks:]
            if nh5 > nb:
                # assumes they are MINC blocks
                blocks += ['     '+str(i) for i in range(nh5 - nb)]
                w_blocks += ['     '+str(i) for i in range(nh5 - nb)]
        else:
            # print('wlisting.element: uses waiwera natural index (as str) as key.')
            blocks = [str(i) for i in range(nh5)]
            w_blocks = blocks
        ### element table
        if 'cell_fields' in self._h5:
            cols = sorted([c for c in self._h5['cell_fields'].keys() if c.startswith('fluid_')])
            table = listingtable(cols, blocks, num_keys=1)
            self._table['element'] = table
        ### connection table
        if 'face_fields' in self._h5:
            cid1 = self._h5['face_cell_1'][:,0]
            cid2 = self._h5['face_cell_2'][:,0]
            # .face_idx equivalent to .cell_idx and .source_idx, which maps waiwera h5 table into
            # geo/mulgrid order, there is no "natural order" of faces in waiwera, so we have to
            # build one here
            self.face_idx = []
            self.face_idx_dir = []
            if self.geo is not None:
                # print('wlisting.connection: tuple of mulgrid block names (str, str) as key.')
                face_keys = self.geo.block_connection_name_list
                # w_boundary is a list of dict, in the order of wjson boundaries, each dict has key
                # of block1, gives name of b2.
                w_boundary = []
                if self.wjson is not None and 'boundaries' in self.wjson:
                    for bd in self.wjson['boundaries']:
                        b2_names = {}
                        for i1 in bd['faces']['cells']:
                            b1 = w_blocks[i1]
                            b2 = '     ' # default empty , overwritten if atmosphere
                            if bd['faces']['normal'] == [0.0, 0.0, 1.0]:
                                # if connect upwards, assume atmospheric, use mulgrid atm block names
                                b1_col = self.geo.column[self.geo.column_name(b1)]
                                b1_lay = self.geo.layer[self.geo.layer_name(b1)]
                                if self.geo.column_surface_layer(b1_col).name == b1_lay.name:
                                    if self.geo.atmosphere_type == 0:
                                        b2 = self.geo.block_name_list[0]
                                    elif self.geo.atmosphere_type == 1:
                                        b2 = self.geo.block_name(self.geo.layerlist[0].name, b1_col.name)
                            b2_names[b1] = b2
                        w_boundary.append(b2_names)
                # this translate waiwera cell ids in (face_cell_1, face_cell_2) into block names
                # w_connection_name_index has keys of (b1 name, b2 name), value is waiwera face index
                # ordered as in h5 file
                w_connection_name_index = {}
                self.w_connection_index_index = {} # useful if given two waiwera cell indices
                for i,(c1,c2) in enumerate(zip(cid1, cid2)):
                    self.w_connection_index_index[(c1,c2)] = i
                    b1 = w_blocks[c1]
                    if c2 < 0:
                        if w_boundary:
                            # face_cell_2 contains the negative of the (1-based) index of the boundaries
                            # specified in json
                            b2 = w_boundary[-c2-1][b1]
                        else:
                            # either wai JSON not loaded OR none are standard atm conne
                            b2 = '     '
                            # simply assumes atm conne if top of the model
                            b1_col = self.geo.column[self.geo.column_name(b1)]
                            b1_lay = self.geo.layer[self.geo.layer_name(b1)]
                            if self.geo.column_surface_layer(b1_col).name == b1_lay.name:
                                if self.geo.atmosphere_type == 0:
                                    b2 = self.geo.block_name_list[0]
                                elif self.geo.atmosphere_type == 1:
                                    b2 = self.geo.block_name(self.geo.layerlist[0].name, b1_col.name)
                    else:
                        b2 = w_blocks[c2]
                    w_connection_name_index[(b1,b2)] = i
                for c in self.geo.block_connection_name_list:
                    # (cell1, cell2) positive means cell1 -> cell2
                    # NOTE this is opposite to T2/AUT2's result convention!
                    if c in w_connection_name_index:
                        self.face_idx.append(w_connection_name_index[c])
                        self.face_idx_dir.append(-1.0)
                    elif c[::-1] in w_connection_name_index:
                        self.face_idx.append(w_connection_name_index[c[::-1]])
                        self.face_idx_dir.append(1.0)
                    else:
                        debugxx = []
                        for key in w_connection_name_index:
                            if c[0] in key:
                                debugxx.append(key)
                        msg = str(debugxx)
                        msg += '\nMulgrid connection name %s not found in Waiwera H5 output.' % str(c)
                        raise Exception(msg)
            else:
                # print('wlisting.connection: tuple of natural cell indices (as str, as str) as key.')
                # keeps whatever order is in h5, NOTE the order is unpredictable
                face_keys = [(str(i), str(j)) for i,j in zip(cid1, cid2)]
                self.face_idx = list(range(len(cid1)))
                self.face_idx_dir = [-1.0] * len(cid1)

            self.face_idx_dir = np.array(self.face_idx_dir)
            cols = sorted([c for c in self._h5['face_fields'].keys() if c.startswith('flux_')])
            table = listingtable(cols, face_keys, num_keys=2, allow_reverse_keys=True)
            self._table['connection'] = table
        ### gener table
        if 'source_fields' in self._h5:
            self.source_name_index = {} # allows either source name or (block name, gener name) as key
            skip_cols = ['source_' + n for n in ['source_index', 'local_source_index', 'natural_cell_index', 'local_cell_index']]
            cols = sorted([c for c in self._h5['source_fields'].keys() if c.startswith('source_') and c not in skip_cols])
            self.source_idx = self._h5['source_index'][:,0]
            source_keys = None
            if self.geo is not None and self.wjson is not None:
                if 'source' in self.wjson and len(self.wjson['source']) == len(self.source_idx):
                    if all(['name' in s for s in self.wjson['source']]) and all(['cell' in s for s in self.wjson['source']]):
                        # each source has a name, each source has a single cell
                        # print('wlisting.generation: detects matching Waiwera input JSON and HDF5 source_fields, use (block name, source name) as key.')
                        cid = [w_blocks[s['cell']] for s in self.wjson['source']]
                        gid = [str(s['name']) for s in self.wjson['source']]
                        source_keys = list(zip(cid, gid))
                        for i,gk in enumerate(source_keys):
                            self.source_name_index[gk] = i
                    for i,s in enumerate(self.wjson['source']):
                        if 'name' in s:
                            self.source_name_index[s['name']] = i
            if source_keys is None:
                # print('wlisting.generation: use source index (as str) as key.')
                # use source index (as str) as key
                source_keys = [str(i) for i in range(len(self.source_idx))]
                table = listingtable(cols, source_keys, num_keys=1)
            else:
                # source_keys is (bname, gname) as in original t2listing
                table = listingtable(cols, list(zip(cid, gid)), num_keys=2)
            self._table['generation'] = table
        # makes tables in self._table accessible as attributes
        for key,table in self._table.items():
            setattr(self, key, table)
        # have to be get first table ready
        self.index = 0

    def history(self, selection, short=False, start_datetime=None):
        """ Returns time histories for specified selection of table type, names
        (or indices) and column names.  This is implemented to be similar to
        t2listing's .history().

        ('e', block name/index, column name)
            for cell_fields, cell can be specified as Waiwera natural cell index
            (int) or mulgrid's block name (str)

        ('c', (c1, c2), column name)
            for face_fields, a face/connection can be specified by a tuple of
            Waiwera's natural cell index (int, int) or mulgrid's block names
            (str, str).

        ('g', g, column name) OR ('g', (b,g), column name)
            for source_fields, a gener/source can be specified by an index
            (int), a source name (str), or a tuple of (block name, source name)
            ((str, str)).  The tuple option is only available if each source has
            a name and each source has a single cell in the Waiwera JSON input.

        short and start_datetime are not implemented at the moment
        """
        if short is True: raise Exception('.history() short=True not implemented yet')
        if start_datetime is not None: raise Exception('.history() start_datetime not implemented yet')
        if isinstance(selection, tuple):
            selection = [selection]
        results = []
        for tbl,b,cname in selection:
            if tbl == 'e':
                if isinstance(b, str):
                    bi = self.geo.block_name_index[b]
                elif isinstance(b, unicode):
                    bi = self.geo.block_name_index[str(b)]
                elif isinstance(b, int):
                    bi = b
                else:
                    raise Exception('.history() block must be an int or str: %s (%s)' % (str(b),str(type(b))))
                if bi < 0:
                    bi = self.geo.num_blocks + bi
                if bi < self.geo.num_atmosphere_blocks:
                    raise Exception('.history() does not support extracting results for atmosphere blocks')
                ### important to convert cell index
                bbi = self.cell_idx[bi-self.geo.num_atmosphere_blocks]
                ys = self._h5['cell_fields'][cname][:,bbi]
                results.append((self.fulltimes, ys))
            elif tbl == 'c':
                if isinstance(b[0], int):
                    # (i1, i2) assume both are integer cell index in Waiwera sense
                    cci = self.w_connection_index_index[b]
                elif isinstance(b[0], str):
                    # (b1, b2) assume both are string block names in mulgrid
                    if self.geo is None:
                        raise Exception('Mulgrid geometry is required if connection tuple is specified by block names.')
                    ci = self.geo.block_connection_name_index[b]
                    cci = self.face_idx[ci]
                ys = self._h5['face_fields'][cname][:,cci]
                results.append((self.fulltimes, ys))
            elif tbl == 'g':
                if isinstance(b, tuple):
                    # (block name, source name) both str
                    gi = self.source_name_index[b]
                elif isinstance(b, str):
                    # single natural source index !! diff from TOUGH2
                    gi = self.source_name_index[b]
                if isinstance(b, int):
                    # directly as source index
                    gi = b
                ggi = self.source_idx[gi]
                ys = self._h5['source_fields'][cname][:,ggi]
                results.append((self.fulltimes, ys))
            else:
                raise Exception('Unsupported .history() selection table type: %s' % tbl)
        if len(results) == 1: results = results[0]
        return results

    def read_tables(self):
        """ copy values from h5 into listingtables, with slicing """
        if 'element' in self.table_names:
            nh5 = len(self.cell_idx)
            for i,cname in enumerate(self.element.column_name):
                self.element._data[-nh5:,i] = self._h5['cell_fields'][cname][self._index][self.cell_idx]
        if 'connection' in self.table_names:
            for i,cname in enumerate(self.connection.column_name):
                # re-order as geo.block_connection_name_list and reverse values if required
                self.connection._data[:,i] = self._h5['face_fields'][cname][self._index][self.face_idx] * self.face_idx_dir
        if 'generation' in self.table_names:
            for i,cname in enumerate(self.generation.column_name):
                self.generation._data[:,i] = self._h5['source_fields'][cname][self._index][self.source_idx]

    def get_index(self): return self._index
    def set_index(self, i):
        self._index = i
        if self._index < 0: self._index += self.num_fulltimes
        self.read_tables()
    index = property(get_index, set_index)

    def first(self): self.index = 0
    def last(self): self.index = -1
    def next(self):
        """Find and read next set of results; returns false if at end of listing"""
        more = self.index < self.num_fulltimes - 1
        if more: self.index += 1
        return more
    def prev(self):
        """Find and read previous set of results; returns false if at start of listing"""
        more = self.index > 0
        if more: self.index -= 1
        return more

    def get_table_names(self):
        return sorted(self._table.keys())
    table_names = property(get_table_names)

    def get_time(self): return self.fulltimes[self.index]
    def set_time(self, t):
        if t < self.fulltimes[0]: self.index = 0
        elif t > self.fulltimes[-1]: self.index = -1
        else:
            dt = np.abs(self.fulltimes - t)
            self.index = np.argmin(dt)
    time = property(get_time, set_time)


class test_medium(unittest.TestCase):
    def setUp(self):
        from mulgrids import mulgrid
        self.geo = mulgrid('g2medium.dat')
        self.lst = wlisting('2DM002.h5', self.geo)

    def test_atm_blocks(self):
        self.assertEqual(len(self.lst.element.row_name), self.geo.num_blocks)
        # atmosphere blocks should be zero
        self.assertEqual(
            list(self.lst.element['fluid_temperature'][:self.geo.num_atmosphere_blocks]),
            [0.0] * self.geo.num_atmosphere_blocks)
        # even after change index
        self.index = 1
        self.assertEqual(
            list(self.lst.element['fluid_temperature'][:self.geo.num_atmosphere_blocks]),
            [0.0] * self.geo.num_atmosphere_blocks)

    def test_tables(self):
        self.assertEqual(self.lst.table_names, ['element', 'generation'])
        cols = [
            'fluid_liquid_capillary_pressure',
            'fluid_liquid_density',
            'fluid_liquid_internal_energy',
            'fluid_liquid_relative_permeability',
            'fluid_liquid_saturation',
            'fluid_liquid_specific_enthalpy',
            'fluid_liquid_viscosity',
            'fluid_liquid_water_mass_fraction',
            'fluid_phases',
            'fluid_pressure',
            'fluid_region',
            'fluid_temperature',
            'fluid_vapour_capillary_pressure',
            'fluid_vapour_density',
            'fluid_vapour_internal_energy',
            'fluid_vapour_relative_permeability',
            'fluid_vapour_saturation',
            'fluid_vapour_specific_enthalpy',
            'fluid_vapour_viscosity',
            'fluid_vapour_water_mass_fraction',
            'fluid_water_partial_pressure']
        self.assertEqual(sorted(self.lst.element.column_name), sorted(cols))
        cols = [
            'source_component',
            'source_enthalpy',
            'source_rate',
            ]
        self.assertEqual(sorted(self.lst.generation.column_name), sorted(cols))

    def test_basic_properties(self):
        self.assertEqual(self.lst.num_fulltimes, 2)
        np.testing.assert_almost_equal(
            self.lst.fulltimes,
            [0.0, 1.0E16],
            decimal=7)

    def test_index(self):
        self.assertEqual(self.lst.index, 0)
        self.assertAlmostEqual(self.lst.time, 0.0)
        np.testing.assert_almost_equal(
            self.lst.element['fluid_temperature'][-5:],
            [15.0, 15.0, 15.0, 15.0, 15.0]
            )

        self.lst.index = 1
        self.lst.index = 1
        self.assertEqual(self.lst.index, 1)
        self.assertAlmostEqual(self.lst.time, 1.0E16)
        np.testing.assert_almost_equal(
            self.lst.element['fluid_temperature'][-5:],
            [235.7365710, 234.266913, 233.142164, 232.376319, 231.98525],
            decimal=4
            )

        self.lst.index = 0
        self.assertEqual(self.lst.index, 0)
        self.assertAlmostEqual(self.lst.time, 0.0)
        np.testing.assert_almost_equal(
            self.lst.element['fluid_temperature'][-5:],
            [15.0, 15.0, 15.0, 15.0, 15.0]
            )

        self.lst.index = -1
        self.assertEqual(self.lst.index, 1)
        self.assertAlmostEqual(self.lst.time, 1.0E16)
        np.testing.assert_almost_equal(
            self.lst.element['fluid_temperature'][-5:],
            [235.7365710, 234.266913, 233.142164, 232.376319, 231.98525],
            decimal=4
            )

        ### generation
        np.testing.assert_almost_equal(
            self.lst.generation['source_rate'][:3],
            [0.075, 0.075, 160.0],
            decimal=4
            )
        np.testing.assert_almost_equal(
            self.lst.generation['source_enthalpy'][:3],
            [1200000.0, 1200000.0, 0.0],
            decimal=4
            )
        np.testing.assert_almost_equal(
            self.lst.generation['source_component'][:3],
            [1.0, 1.0, 2.0],
            decimal=4
            )

    def test_time(self):
        self.lst.time = self.lst.fulltimes[1] - 100.0
        self.assertEqual(self.lst.index, 1)
        self.lst.time = 1.0e19
        self.assertEqual(self.lst.index, 1)
        self.lst.time = 0.0
        self.assertEqual(self.lst.index, 0)
        self.lst.time = 100.0
        self.assertEqual(self.lst.index, 0)

    def test_history(self):
        # use relative tolerance, expect minor diff with different num of cpus
        rtol = 1e-10
        xs, ys = self.lst.history(('e', -1, 'fluid_pressure'))
        np.testing.assert_allclose(xs, [0, 1.0e16], rtol=rtol)
        np.testing.assert_allclose(ys, [101350.0, 1.3010923804323431E7], rtol=rtol)
        xs, ys = self.lst.history(('e', 339, 'fluid_pressure'))
        np.testing.assert_allclose(xs, [0, 1.0e16], rtol=rtol)
        np.testing.assert_allclose(ys, [101350.0, 1.3010923804323431E7], rtol=rtol)
        xs, ys = self.lst.history(('e', '  t16', 'fluid_pressure'))
        np.testing.assert_allclose(xs, [0, 1.0e16], rtol=rtol)
        np.testing.assert_allclose(ys, [101350.0, 1.3010923804323431E7], rtol=rtol)
        xs, ys = self.lst.history(('e', 338, 'fluid_temperature'))
        np.testing.assert_allclose(xs, [0, 1.0e16], rtol=rtol)
        np.testing.assert_allclose(ys, [15.0, 232.3763193900396], rtol=rtol)
        # cell index doesn't matter in generation
        xs, ys = self.lst.history(('g', (999, 0), 'source_enthalpy'))
        np.testing.assert_allclose(xs, [0, 1.0e16], rtol=rtol)
        np.testing.assert_allclose(ys, [1200e3, 1200e3], rtol=rtol)
        # also accepts single gener index int
        xs, ys = self.lst.history(('g', 0, 'source_enthalpy'))
        np.testing.assert_allclose(xs, [0, 1.0e16], rtol=rtol)
        np.testing.assert_allclose(ys, [1200e3, 1200e3], rtol=rtol)
        # also accepts gener index as str
        xs, ys = self.lst.history(('g', '0', 'source_enthalpy'))
        np.testing.assert_allclose(xs, [0, 1.0e16], rtol=rtol)
        np.testing.assert_allclose(ys, [1200e3, 1200e3], rtol=rtol)
        # cell index doesn't matter in generation
        xs, ys = self.lst.history(('g', (999, '0'), 'source_enthalpy'))
        np.testing.assert_allclose(xs, [0, 1.0e16], rtol=rtol)
        np.testing.assert_allclose(ys, [1200e3, 1200e3], rtol=rtol)
        tbl = self.lst.history([
            ('e', -1, 'fluid_pressure'),
            ('e', 339, 'fluid_pressure'),
            ('e', '  t16', 'fluid_pressure'),
            ('e', 338, 'fluid_temperature'),
            ('g', (999, 0), 'source_enthalpy'),
            ])
        np.testing.assert_allclose(tbl[0][0], [0, 1.0e16], rtol=rtol)
        np.testing.assert_allclose(tbl[0][1], [101350.0, 1.3010923804323431E7], rtol=rtol)
        np.testing.assert_allclose(tbl[1][1], [101350.0, 1.3010923804323431E7], rtol=rtol)
        np.testing.assert_allclose(tbl[2][1], [101350.0, 1.3010923804323431E7], rtol=rtol)
        np.testing.assert_allclose(tbl[3][1], [15.0, 232.3763193900396], rtol=rtol)
        np.testing.assert_allclose(tbl[4][1], [1200e3, 1200e3], rtol=rtol)

        # should spit out an exception about not supporting atmosphere blocks
        self.assertRaises(Exception, self.lst.history, ('e', 0, 'fluid_temperature'))
        self.assertRaisesRegexp(Exception, 'atmosphere', self.lst.history, ('e', 0, 'fluid_temperature'))

class test_medium_multiple_cpu(test_medium):
    def setUp(self):
        from mulgrids import mulgrid
        self.geo = mulgrid('g2medium.dat')
        self.lst = wlisting('2DM002a.h5', self.geo)

class test_compare(unittest.TestCase):
    def setUp(self):
        from mulgrids import mulgrid
        self.geo = mulgrid('g2medium.dat')
        self.lst1 = wlisting('2DM002.h5', self.geo)
        self.lst2 = wlisting('2DM002a.h5', self.geo)

    def test_table(self):
        self.lst1.index = 1
        self.lst2.index = 1
        self.assertAlmostEqual(self.lst1.time, 1.0E16)
        self.assertAlmostEqual(self.lst2.time, 1.0E16)
        np.testing.assert_allclose(
            self.lst1.element['fluid_temperature'],
            self.lst2.element['fluid_temperature'],
            rtol=1e-10,
            equal_nan=True
            )
        np.testing.assert_allclose(
            self.lst1.element['fluid_pressure'],
            self.lst2.element['fluid_pressure'],
            rtol=1e-10,
            equal_nan=True
            )

    def test_history(self):
        np.testing.assert_allclose(
            self.lst1.history(('e', -1, 'fluid_pressure'))[1],
            self.lst2.history(('e', -1, 'fluid_pressure'))[1],
            rtol=1e-10,
            )
        np.testing.assert_allclose(
            self.lst1.history(('e', 128, 'fluid_temperature'))[1],
            self.lst2.history(('e', 128, 'fluid_temperature'))[1],
            rtol=1e-10,
            )

        # check for false positive, this should be different
        self.assertRaises(
            AssertionError,
            np.testing.assert_allclose,
            self.lst1.history(('e', 123, 'fluid_pressure'))[1],
            self.lst2.history(('e', 78, 'fluid_pressure'))[1],
            rtol=1e-10,
            )
        self.assertRaises(
            AssertionError,
            np.testing.assert_allclose,
            self.lst1.history(('e', 127, 'fluid_temperature'))[1],
            self.lst2.history(('e', 128, 'fluid_temperature'))[1],
            rtol=1e-10,
            )



if __name__ == '__main__':
    unittest.main(verbosity=2)

    # import time
    # from mulgrids import mulgrid
    # geo = mulgrid('gLihir_v7_NS.dat')
    # init_time = time.time()
    # lst = wlisting('Lihir_v7_SP_NS_060_wai.h5', geo)
    # print '%.2f sec' % (time.time() - init_time)
    # init_time = time.time()
    # lst.index = 1
    # print '%.2f sec' % (time.time() - init_time)






"""Definition file for parameter data types"""
#
# each function must accept three params: name,dat,val
#     names: t2data object names, eg. block, gener etc, must be a list
#     dat: your normal t2data object from PyTOUGH
#     val: a number that is going to be assigned, when not exist value
#          from dat will be returned, the first in the name list is used
#
# use existing entries as example to write new ones
# your new function name can then be used as 'param_type'
#
# the shortNames needs to be filled in, the right hand side
# string must be EXACTLY 3 characters wide, and NO SPACE
# these short names will be used as part of names in * parameter data
#

# TODO: consider using munch to allow access dictionary with object-oriented
# style? see:
# https://github.com/Infinidat/munch
# https://stackoverflow.com/questions/1305532/how-to-convert-a-nested-python-dict-to-object

import re

from gopest.common import getFromDict, setInDict

class ParDef(object):
    """ Base class for parameter, used to access model input parameters """
    def __init__(self, simulator):
        super(ParDef, self).__init__()
        if simulator not in ['aut2', 'waiwera']:
            raise Exception("simulator '%s' not supported." % simulator)
        self.simulator = simulator

    def get(self, dat, name):
        _get = getattr(self, 'get_' + self.simulator)
        return _get(dat, name)
    def get_aut2(self, dat, name):
        raise NotImplementedError(self.__class__.__name__)
    def get_waiwera(self, dat, name):
        raise NotImplementedError(self.__class__.__name__)

    def set(self, dat, name, value):
        _set = getattr(self, 'set_' + self.simulator)
        return _set(dat, name, value)
    def set_aut2(self, dat, name, value):
        raise NotImplementedError(self.__class__.__name__)
    def set_waiwera(self, dat, name, value):
        raise NotImplementedError(self.__class__.__name__)

    def find_names(self, dat, pattern):
        _find_names = getattr(self, 'find_names_' + self.simulator)
        return _find_names(dat, pattern)
    def find_names_aut2(self, dat, pattern):
        raise NotImplementedError(self.__class__.__name__)
    def find_names_waiwera(self, dat, pattern):
        raise NotImplementedError(self.__class__.__name__)

class permeability_1_byrock(ParDef):
    def get_aut2(self, dat, name):
        return dat.grid.rocktype[name].permeability[0]

    def set_aut2(self, dat, name, value):
        dat.grid.rocktype[name].permeability[0] = value

    def find_names_aut2(self, dat, pattern):
        pattern = re.compile(name)
        return [r.name for r in dat.grid.rocktypelist if pattern.match(r.name)]

    def get_waiwera(self, dat, name):
        for rock in dat['rock']['types']:
            if rock['name'] == name:
                return rock['permeability'][0]

    def set_waiwera(self, dat, name, value):
        for rock in dat['rock']['types']:
            if rock['name'] == name:
                rock['permeability'][0] = value

    def find_names_waiwera(self, dat, pattern):
        rex = re.compile(pattern)
        return [r['name'] for r in dat['rock']['types'] if rex.match(r['name'])]

class permeability_2_byrock(ParDef):
    def get_aut2(self, dat, name):
        return dat.grid.rocktype[name].permeability[1]

    def set_aut2(self, dat, name, value):
        dat.grid.rocktype[name].permeability[1] = value

    def find_names_aut2(self, dat, pattern):
        pattern = re.compile(name)
        return [r.name for r in dat.grid.rocktypelist if pattern.match(r.name)]

    def get_waiwera(self, dat, name):
        for rock in dat['rock']['types']:
            if rock['name'] == name:
                return rock['permeability'][1]

    def set_waiwera(self, dat, name, value):
        for rock in dat['rock']['types']:
            if rock['name'] == name:
                rock['permeability'][1] = value

    def find_names_waiwera(self, dat, pattern):
        rex = re.compile(pattern)
        return [r['name'] for r in dat['rock']['types'] if rex.match(r['name'])]

class permeability_3_byrock(ParDef):
    def get_aut2(self, dat, name):
        return dat.grid.rocktype[name].permeability[2]

    def set_aut2(self, dat, name, value):
        dat.grid.rocktype[name].permeability[2] = value

    def find_names_aut2(self, dat, pattern):
        pattern = re.compile(name)
        return [r.name for r in dat.grid.rocktypelist if pattern.match(r.name)]

    def get_waiwera(self, dat, name):
        for rock in dat['rock']['types']:
            if rock['name'] == name:
                return rock['permeability'][2]

    def set_waiwera(self, dat, name, value):
        for rock in dat['rock']['types']:
            if rock['name'] == name:
                rock['permeability'][2] = value

    def find_names_waiwera(self, dat, pattern):
        rex = re.compile(pattern)
        return [r['name'] for r in dat['rock']['types'] if rex.match(r['name'])]

class porosity_byrock(ParDef):
    def get_aut2(self, dat, name):
        return dat.grid.rocktype[name].porosity

    def set_aut2(self, dat, name, value):
        dat.grid.rocktype[name].porosity = value

    def find_names_aut2(self, dat, pattern):
        pattern = re.compile(name)
        return [r.name for r in dat.grid.rocktypelist if pattern.match(r.name)]

    def get_waiwera(self, dat, name):
        for rock in dat['rock']['types']:
            if rock['name'] == name:
                return rock['porosity']

    def set_waiwera(self, dat, name, value):
        for rock in dat['rock']['types']:
            if rock['name'] == name:
                rock['porosity'] = value

    def find_names_waiwera(self, dat, pattern):
        rex = re.compile(pattern)
        return [r['name'] for r in dat['rock']['types'] if rex.match(r['name'])]

class massgener_rate(ParDef):
    def get_aut2(self, dat, name):
        for g in dat.generatorlist:
            if g.name == name:
                return g.gx

    def set_aut2(self, dat, name, value):
        for g in dat.generatorlist:
            if g.name == name:
                g.gx = value
                return

    def find_names_aut2(self, dat, pattern):
        pattern = re.compile(name)
        return [g.name for g in dat.generatorlist if pattern.match(g.name)]

    def get_waiwera(self, dat, name):
        for source in dat['source']:
            if source['name'] == name:
                return source['rate']

    def set_waiwera(self, dat, name, value):
        for source in dat['source']:
            if source['name'] == name:
                source['rate'] = value

    def find_names_waiwera(self, dat, pattern):
        rex = re.compile(pattern)
        return [s['name'] for s in dat['source'] if rex.match(s['name'])]


class json_values(ParDef):
    def get_aut2(self, dat, name):
        """ expects name to be a tuple ([k1, k2, k3], last_key), where a list of
        keys k1,k2,k3 (can use either str or int) leading up to just before the
        last key.
        """
        cfg = getFromDict(dat.config, name[0])
        return cfg[name[1]]

    def set_aut2(self, dat, name, value):
        cfg = getFromDict(dat.config, name[0])
        cfg[name[1]] = value

    def find_names_aut2(self, dat, pattern):
        rex = re.compile(name)
        cfg = getFromDict(dat.config, name[0])
        return [(name[0], z) for z in cfg.keys() if rex.match(z)]


shortNames = {
'CJ' : 'json_values',
'HR' : 'heatgener_rate',
'SR' : 'massgener_rate',
'SE' : 'massgener_enth',
'IN' : 'infiltration',
'K1' : 'permeability_1',
'K2' : 'permeability_2',
'K3' : 'permeability_3',
'PO' : 'porosity',
'R1' : 'permeability_1_byrock',
'R2' : 'permeability_2_byrock',
'R3' : 'permeability_3_byrock',
'RA' : 'permeability_123_byrock',  #tgra963
'RS' : 'permeability_12_byrock',   #tgra963
'RP' : 'porosity_byrock',
'RE' : 'upflow_rech'
}


def upflow_rech(names, dat, val=None):
    if val is None:
        return dat.config['RechCoefficients'][names[0]]
    else:
        dat.config['RechCoefficients'][names[0]] = val

def heatgener_rate(names,dat,val=None):
    if val is None:
        for g in dat.generatorlist:
            if g.name in names:
                return g.gx
    else:
        for g in dat.generatorlist:
            if g.name in names:
                g.gx = val

def massgener_enth(names,dat,val=None):
    if val is None:
        for g in dat.generatorlist:
            if g.name in names:
                return g.ex
    else:
        for g in dat.generatorlist:
            if g.name in names:
                g.ex = val

def infiltration(names,dat,val=None):
    if val is None:
        return 999 # yet to be done
    else:
        # doesn't care names
        from make_rain import make_rain
        cfg_name = 'make_rain.cfg'
        cfg_name = os.getcwd() + os.path.sep + cfg_name
        cfg = config(cfg_name)
        [annualrain_const, annualrain_history,
            infiltration, raintemp, newwell_label, tim2sec, cols] = make_rain_proc_cfg(cfg)

        infiltration = val

        total_rain = make_rain(geo,dat,annualrain_const,infiltration,raintemp,newwell_label,
            60.0*60.0*24.0*365.25,cols)

def permeability_1(names,dat,val=None):
    if val is None:
        return dat.grid.block[names[0]].rocktype.permeability[0]
    else:
        from lib_rocktypes import update_rocktype_property_byblocks
        update_rocktype_property_byblocks('permeability[0]',names,dat,val)

def permeability_2(names,dat,val=None):
    if val is None:
        return dat.grid.block[names[0]].rocktype.permeability[1]
    else:
        from lib_rocktypes import update_rocktype_property_byblocks
        update_rocktype_property_byblocks('permeability[1]',names,dat,val)

def permeability_3(names,dat,val=None):
    if val is None:
        return dat.grid.block[names[0]].rocktype.permeability[2]
    else:
        from lib_rocktypes import update_rocktype_property_byblocks
        update_rocktype_property_byblocks('permeability[2]',names,dat,val)

def porosity(names,dat,val=None):
    if val is None:
        return dat.grid.block[names[0]].rocktype.porosity
    else:
        from lib_rocktypes import update_rocktype_property_byblocks
        update_rocktype_property_byblocks('porosity',names,dat,val)

def permeability_123_byrock(names,dat,val=None):
    if val is None:
        return dat.grid.rocktype[names[0]].permeability[0]
    else:
        dat.grid.rocktype[names[0]].permeability[0] = val
        dat.grid.rocktype[names[0]].permeability[1] = val
        dat.grid.rocktype[names[0]].permeability[2] = val

def permeability_12_byrock(names,dat,val=None):
    if val is None:
        return dat.grid.rocktype[names[0]].permeability[0]
    else:
        dat.grid.rocktype[names[0]].permeability[0] = val
        dat.grid.rocktype[names[0]].permeability[1] = val




import json
from os.path import splitext, basename, isfile
import inspect
import functools
from collections import OrderedDict

from t2data import *

from gopest.common import TwoWayDict
from gopest.common import Singleton
from gopest.common import readList
from gopest.common import updateObj
from gopest.common import private_cleanup_name

# implementation:
#   a user entry is something simple a user can undersand and easily enter
#   these are entered as groups, each with type and settings.  these are
#   turned into what PEST requires.
#

# i need a dictionary, leys of parameter data type (short), which maps to
# functions from the user defined functions that can modify t2data objects

from gopest.common import config as cfg
INPUT_TYPE = cfg['simulator']['input-type']

from gopest import par_def
par_classes = {}
for n,c in inspect.getmembers(par_def, inspect.isclass):
    if issubclass(c, par_def.ParDef):
        par_classes[n] = c

class PestParamGroups(object):
    """ representation of entries in PEST * parameter groups """
    def __init__(self,name,INCTYP='relative',DERINC=0.01,DERINCLB=0.0,
        FORCEN='switch',DERINCMUL=2.0,DERMTHD='parabolic'):
        self.PARGPNME = name
        self.INCTYP = INCTYP
        self.DERINC=DERINC
        self.DERINCLB=DERINCLB
        self.FORCEN=FORCEN
        self.DERINCMUL=DERINCMUL
        self.DERMTHD=DERINCMUL

        self.defPARLBND = 0.0
        self.defPARUBND = 0.0

class PestParamData(object):
    """ this is equalivalent to each of * parameter data in PEST.  it is
    able to write entries for PEST control file (.pst), and PEST template
    file (.tpl).  It is also responsible of reading PEST generated 'model'
    file and make real changes into Tough2 input file """
    #? do I have too many tasks for this class?
    def __init__(self,name='',PARTRANS='none',PARCHGLIM='factor',PARVAL1=1.0,
        PARLBND=0.5,PARUBND=1.5,PARGP='grp',SCALE=1.0,OFFSET=0.0,DERCOM=1,
        TEMPLATEWIDTH=7):
        self.PARNME = name
        self.PARTRANS=PARTRANS
        self.PARCHGLIM=PARCHGLIM
        self.PARVAL1=PARVAL1
        self.PARLBND=PARLBND
        self.PARUBND=PARUBND
        self.PARGP=PARGP
        self.SCALE=SCALE
        self.OFFSET=OFFSET
        self.DERCOM=DERCOM

        # this is not part of PEST's design, used by goPEST to determine the
        # correct template field width, which is related to the precision of
        # TOUGH2 input.  It should be 7 for a standard TOUGH2 entry, with 2
        # less for starting- and ending-mark, and one less because PEST has
        # better ability to squeeze one extra precision than Python using the
        # same width.  If using .pdat, you should probably use 12 (15-3) to get
        # the precision effect.
        self.TEMPLATEWIDTH = TEMPLATEWIDTH

    def write(self, f=None):
        """ expects f to be a file object ready to write(), otherwise print to screen """
        line = ' '.join([str(a) for a in [
            self.PARNME,
            self.PARTRANS,
            self.PARCHGLIM,
            self.PARVAL1,
            self.PARLBND,
            self.PARUBND,
            self.PARGP,
            self.SCALE,
            self.OFFSET,
            self.DERCOM,
            ]])
        if f is None:
            print(line)
        else:
            f.write(line + '\n')


PARAM_ALIAS = TwoWayDict(par_def.shortNames)

class PestParamDataName(Singleton):
    """ remembers a list of parameter data and observation data """
    def __init__(self):
        self.names = set([])
    def newName(self,basename):
        apnd = ' 01234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
        i = 0
        newname = (basename + apnd[i]).strip()
        while newname in self.names:
            i += 1
            if i >= len(apnd):
                raise Exception('Unable to find unused parameter name for:' + basename)
                break
            newname = (basename + apnd[i]).strip()
        return newname
    def add(self,newname):
        self.names.add(newname)

class UserEntryParam(object):
    """ understands user entry group and its setings, type.  and will be
    responsible of creating PestParamData objects, as each * parameter data
    in PEST """
    ListType = ['ForEach','SingleValue','TiedValues']
    def __init__(self, paramListType, paramList, t2objListType, t2objList, t2objType,
                 dat, defaults=None):
        # paramListType  'ForEach' or 'SingleValue' or 'TiedValues'
        # paramList      a list of 'permeability_1' or 'porosity' etc
        # t2objListType  'ForEach' or 'SingleValue' or 'TiedValues'
        # t2objList      a list of PyTOUGH objects names 'aaa15' or 'BA***' etc
        # t2objType      one of 'Rocktype', 'Element', or 'Generator'
        # defaults       dict of default PestParamData of each PARAM_TYPE
        if paramListType not in self.ListType: raise Exception(paramListType, ' is not in ', ListType)
        if t2objListType not in self.ListType: raise Exception(t2objListType, ' is not in ', ListType)
        for p in [p for p in paramList if p not in par_classes]: raise Exception(p, ' is not valid')
        self.t2objList = t2objList
        self.t2objType = t2objType
        self.paramListType=paramListType
        self.paramList=paramList
        self.t2objListType=t2objListType
        from copy import deepcopy
        self.defaults=deepcopy(defaults)

        # generate these by calling makeParamDataNames, pest *  parameter data entries
        self.paramData = None
        self.ties = None

    def makeParamData(self,dat,tpl):
        """ generate a list of paramData and write to PEST template file"""
        # ??? Maybe I should make it spit tpl file entries, instead of write directly into file?
        if (self.paramData is not None) and (self.ties is not None): return
        self.paramData = []
        self.ties = []

        # reuse ParDef instances (used as getter/setter here) for param type
        par_getters = {}
        for pt in self.paramList:
            if pt not in par_getters:
                par_getters[pt] = par_classes[pt](INPUT_TYPE)

        # obtain a list of names that exists in all self.paramList and self.t2objList
        # use OrderedDict as an ordered set, to remember order (ordr is noce to have)
        names = []
        intersect = None
        for pt in self.paramList:
            for pattern in self.t2objList:
                a = par_getters[pt].find_names(dat, pattern)
                names += a
                if intersect is None:
                    intersect = set(a)
                else:
                    intersect = intersect & set(a)
        name_order = OrderedDict.fromkeys(names).keys()
        matchedList = [n for n in name_order if n in intersect]

        # if both ForEach
        for t2o in matchedList:
            for pt in self.paramList:
                if (pt in self.defaults) and (len(self.defaults[pt].PARNME) == 2):
                    # if set properly already (len=2), use the specified value
                    if isinstance(t2o, tuple):
                        nname = PestParamDataName().newName(self.defaults[pt].PARNME + private_cleanup_name(t2o[-1][:5]))
                    else:
                        nname = PestParamDataName().newName(self.defaults[pt].PARNME + private_cleanup_name(t2o))
                else:
                    # use PARAM_ALIAS as name base if not set properly
                    if isinstance(t2o, tuple):
                        nname = PestParamDataName().newName(PARAM_ALIAS[pt] + private_cleanup_name(t2o[-1][:5]))
                    else:
                        nname = PestParamDataName().newName(PARAM_ALIAS[pt] + private_cleanup_name(t2o))
                PestParamDataName().add(nname)
                if self.defaults is not None:
                    if pt in self.defaults:
                        from copy import deepcopy
                        pd = deepcopy(self.defaults[pt])
                    else:
                        pd = PestParamData()
                else:
                    pd = PestParamData()
                pd.PARNME = nname
                pd.PARVAL1 = par_getters[pt].get(dat, t2o)
                self.paramData.append(pd)
                theline = "$%-" + str(pd.TEMPLATEWIDTH) + 's$, %-20s, "%s"\n'
                tpl.write(theline % (nname, ('"%s"' % pt), str([t2o])))

def readUserParameter(userListName, dat):
    """ returns a list of UserEntryParam from reading the file with name
        userListName """
    userEntries = []
    with open(userListName,'r') as f:
        entryName, entry = readList(f)
    # some defaults even if no sections exists:
    parDefaults = {}
    for i,en in enumerate(entryName):
        if en == 'Defaults':
            for line in entry[i]:
                if ':' in line:
                    ptype = [s.strip() for s in line.split(':')[0].split(',')]
                    for p in ptype:
                        if p not in parDefaults: parDefaults[p] = PestParamData()
                else:
                    for p in ptype:
                        updateObj(parDefaults[p],[line])
        if en == 'Param':
            if len(entry[i]) < 2: raise Exception
            tmp1, tmp2 = entry[i][0].split(':')
            paramListType = tmp1.strip()
            paramList = [s.strip() for s in tmp2.split(',')]
            tmp1, tmp2 = entry[i][1].split(':')
            t2objListType = tmp1.strip()
            t2objType = tmp2.strip()
            t2objList = []
            for line in entry[i][2:]:
                t2objList.append(eval(line))
            userEntries.append(UserEntryParam(paramListType,paramList,
                t2objListType,t2objList,t2objType,dat,parDefaults))
    return userEntries

def load_model_config(dat):
    """ Return config dict loaded from .json file with same model base name.
    """
    jname = splitext(dat.filename)[0] + '.json'
    # print("*****", jname)
    if isfile(jname):
        with open(jname, 'rU') as jf:
            return json.load(jf)
    else:
        return {}

def save_model_config(dat, config):
    """ Saves config (dict) object into .json using same basename as t2data.
    """
    jname = splitext(dat.filename)[0] + '.json'
    with open(jname, 'w') as jf:
        json.dump(config, jf, indent=4)

def generate_params_and_tpl(origInput, tplToWrite, par_data):
    """ this reads goPESTpar.list and generate appropriate template file and
    writes * parameter data lines into a file """
    if INPUT_TYPE == 'aut2':
        dat = t2data(origInput)
        dat.config = load_model_config(dat)
    elif INPUT_TYPE == 'waiwera':
        with open(origInput, 'r') as f:
            dat = json.load(f)
    else:
        raise Exception()

    uentry = readUserParameter('goPESTpar.list', dat)

    if par_data is not None:
        parf = open(par_data, 'w')
    else:
        parf = None
    tpl = open(tplToWrite, 'w')
    tpl.write('ptf $\n')
    for up in uentry:
        up.makeParamData(dat,tpl)
        for pp in up.paramData:
            pp.write(parf)
    tpl.close()
    if parf is not None:
        parf.close()

def generate_real_model(origInput, pestModel, realInput):
    """ this reads PEST generated model file and create the real TOUGH2 model
    """
    if INPUT_TYPE == 'aut2':
        dat = t2data(origInput)
        dat.config = load_model_config(dat)
    elif INPUT_TYPE == 'waiwera':
        with open(origInput, 'r') as f:
            dat = json.load(f)
    else:
        raise Exception()

    # reuse ParDef instances (used as getter/setter here) for param type
    par_setters = {}

    with open(pestModel,'r') as pmodel:
        for line in pmodel.readlines():
            vals = eval(line.strip())
            pestValue, paramType, names = vals[0], vals[1], eval(vals[2])
            if paramType not in par_setters:
                par_setters[paramType] = par_classes[paramType](INPUT_TYPE)
            par_setters[paramType].set(dat, names[0], pestValue)

    if INPUT_TYPE == 'aut2':
        dat.write(realInput, extra_precision=True, echo_extra_precision=True)
        save_model_config(dat, dat.config)
    elif INPUT_TYPE == 'waiwera':
        with open(realInput, 'w') as f:
            json.dump(dat, f, indent=4, sort_keys=True)

def goPESTpar(argv=[]):
    userlistname = 'goPESTpar.list'

    if len(argv) not in [3,4]:
        print('to generate PEST .pst sections and .tpl (for once): ')
        print('     gopest par origINPUT newPESTtpl')
        print('to read PEST generated model file and create real INPUT for Tough2 (for each PEST forward run):')
        print('     gopest par origINPUT pestINPUT realINPUT')

    if len(argv) == 3:
        origInput = argv[1]
        tplToWrite = argv[2]
        # write *param data lines to screen
        generate_params_and_tpl(origInput, tplToWrite, None)

    if len(argv) == 4:
        origInput = argv[1]
        pestModel = argv[2]
        realInput = argv[3]

        generate_real_model(origInput, pestModel, realInput)



from gopest.par import generate_params_and_tpl
from gopest.obs import generate_obses_and_ins

from gopest.common import config
from gopest.common import runtime

import os
import re
import shutil

def replace_section(begin_sec, end_sec, pcf_text, repl):
    """ returns PEST control file (passed in as a multiline string) with section
    between begin_sec and end_sec replaced by string repl.  It does not matter
    if repl has line breaks at the end or start, they will be cleaned up so the
    final file has no empty lines. """
    pattern = re.compile(r'\n *%s *\n.*?\n *%s *\n' % (re.escape(begin_sec), re.escape(end_sec)), flags=re.DOTALL)
    repl = '\n%s\n%s\n%s\n' % (begin_sec, repl.strip("\n"), end_sec)
    result, cnt = pattern.subn(repl, pcf_text)
    if cnt < 1:
        raise Exception("unable to find loaction between section %s and %s" % (begin_sec, end_sec))
    return result

def get_lines(filename):
    """ return all non-empty lines froma file as a single string, with a line
    count. """
    f = open(filename, 'rU')
    lines, cnt = '', 0
    for line in f:
        if line.strip():
            cnt += 1
            lines += line
    f.close()
    return lines, cnt

def replace_nth_line(longstring, i, repl):
    """ replace ith line with repl, if repl is a function, it will be called to
    process the original line. """
    line_i = re.compile(r'([^\n]*\n){%i}' % i, flags=re.MULTILINE)
    m = line_i.match(longstring)

    if m:
        if isinstance(repl, str):
            real_repl = repl
        elif hasattr(repl, '__call__'):
            real_repl = repl(m.group(1))
        else:
            # if not a string or callable, try convert it into string
            real_repl = str(repl)
        return longstring[:m.start(1)] + real_repl.strip() + '\n' + longstring[m.end(1):]
    else:
        raise Exception("replace_nth_line() failed to replace line number %i" % i)

#############################################################################

def fixpcf_modelcmd(fpst):
    """ update PEST case file (.pst) with model command line and distribution
    files
    """
    fsave = runtime['filename']['save']
    fincon = runtime['filename']['incon']
    fdatns = runtime['filename']['dat_seq'][0]
    flstpr = runtime['filename']['lst_seq'][-1]

    fpst_bk = fpst + '.backup'
    if not os.path.isfile(fpst):
        print('Error: %s does not exist.' % fpst)
        exit(1)
    if os.path.isfile(fpst_bk):
        os.remove(fpst_bk)
    os.rename(fpst, fpst_bk)

    try:
        with open(fpst_bk,'r') as fin:
            pcf_text = fin.read()

        model_cmd = 'gopest run-pest-model\n'
        model_inout = "\n".join([
            'pest_model.tpl pest_model.dat',
            'pest_model.ins  pest_model.obf',
        ])
        filedist = "\n".join([
            '2 %s %s %s %s' % (fsave, fincon, fincon, fincon),
            '1 %s %s.999' % (fsave, fincon),
            '1 %s %s.999' % (fdatns, fdatns),
            '1 %s %s.999' % (flstpr, flstpr),
            '1 pest_model.dat pest_model.dat.999',
            'command = "gopest save-iter-model"',
        ])

        pcf_text = replace_section("* model command line", "* model input/output",
            pcf_text, model_cmd)
        pcf_text = replace_section("* model input/output", "* prior information",
            pcf_text, model_inout)
        pcf_text = replace_section("* distribution files", "# end",
            pcf_text, filedist)

        with open(fpst, 'w') as fout:
            fout.write(pcf_text)
        print('+++ PEST case control file edited, original file saved as %s' % fpst_bk)
    except Exception as e:
        print(e)
        print('update_case_pst.py unable to proceed, restoring.')
        os.rename(fpst_bk, fpst)

def fixpcf_parobs(fpst, dopar=True, doobs=True):
    """ update PEST case file (.pst) with parameter and observation data
    """
    if dopar is False and doobs is False:
        return

    fpst_bk = fpst + '.backup'
    if not os.path.isfile(fpst):
        print('Error: %s does not exist.' % fpst)
        exit(1)
    if os.path.isfile(fpst_bk):
        os.remove(fpst_bk)
    os.rename(fpst, fpst_bk)

    try:
        with open(fpst_bk,'r') as fin:
            pcf_text = fin.read()

        if dopar:
            par_data, n_par = get_lines('.pest_par_data')
            pcf_text = replace_section("* parameter data", "* observation groups",
                pcf_text, par_data)
            print('+++ found %i parameters' % n_par)

            # replace the count of parameters
            def replace_par_cnts(orig):
                """ first two number is n_par and n_obs """
                nums = orig.split()
                return ' '.join([str(n_par)] + nums[1:])
            pcf_text = replace_nth_line(pcf_text, 4, replace_par_cnts)

        if doobs:
            obs_data, n_obs = get_lines('.pest_obs_data')
            pcf_text = replace_section("* observation data", "* model command line",
                pcf_text, obs_data)
            print('+++ found %i observations' % n_obs)

            # replace the count of parameters
            def replace_obs_cnts(orig):
                """ first two number is n_par and n_obs """
                nums = orig.split()
                return ' '.join(nums[:1] + [str(n_obs)] + nums[2:])
            pcf_text = replace_nth_line(pcf_text, 4, replace_obs_cnts)

        with open(fpst, 'w') as fout:
            fout.write(pcf_text)
        print('+++ PEST case control file edited, original file saved as %s' % fpst_bk)
    except Exception as e:
        print(e)
        print('update_case_pst.py unable to proceed, restoring.')
        os.rename(fpst_bk, fpst)

def copy_model_files():
    """ user specifies model's original files in [model.original]section
    These files will be copied to the working directory, with goPEST's internal
    naming convention.
    """
    def copy_to_cwd(filename, newbase):
        """ copy to working dir and rename, but keeping all extention (to lower case) """
        newname = newbase + os.path.splitext(filename)[1].lower()
        print("  copy '%s' -> '%s'" % (filename, newname))
        shutil.copy2(filename, newname)

    for f in config['model']['original']['geometry-files']:
        copy_to_cwd(f, './g_real_model')
    copy_to_cwd(config['model']['original']['incon-file'], './real_model_incon')

    # copy input and output files for the sequence of 'ns', 'pr', etc
    sequence = config['model']['sequence']
    for seq in sequence:
        copy_to_cwd(config['model']['original']['%s-input-file' % seq], './real_model_%s' % seq)
        if '%s-output-file' % seq in config['model']['original']:
            copy_to_cwd(config['model']['original']['%s-output-file' % seq], './real_model_%s' % seq)
    # make a copy of first dat to keep as original (handy for goPESTpar etc)
    copy_to_cwd(config['model']['original']['%s-input-file' % sequence[0]], './real_model_original')

def make_case_cli(argv=[]):
    """ runs goPEST to set up par and obs entries """
    for a in argv[1:]:
        if a not in ['--no-copy', '--no-par', '--no-obs']:
            raise Exception('Unrecognised option "%s".' % a)
    if '--no-copy' in argv:
        print('+++ use existing model files')
    else:
        print('+++ copy from original model files')
        copy_model_files()

    fgeo = runtime['filename']['geom']
    fdato = runtime['filename']['dat_orig']
    fdats = runtime['filename']['dat_seq']

    dopar = False
    if '--no-par' not in argv:
        print('+++ running goPEST to get par')
        print('  gopestpar', fdato, 'pest_model.tpl', '.pest_par_data')
        generate_params_and_tpl(fdato, 'pest_model.tpl', '.pest_par_data')
        dopar = True

    doobs = False
    if '--no-obs' not in argv:
        print('+++ running goPEST to get obs')
        print('  gopestobs', fgeo, fdats[-1], 'pest_model.ins', '.pest_obs_data')
        generate_obses_and_ins(fgeo, fdats[-1], 'pest_model.ins', '.pest_obs_data')
        doobs = True

    # unfortunately I need to use 'real_model_original_pr.dat' here because it
    # has many GENERs that may not exist in natural state, while still being
    # needed in observations, eg. Production gener's block, hopefully this is
    # okay because we usually don't need to get any actual values out of the
    # real_model_original_pr.dat model.

    # edit them into PEST case control file
    fpst = config['pest']['case-name'] + '.pst'
    fixpcf_parobs(fpst, dopar=dopar, doobs=doobs)
    fixpcf_modelcmd(fpst)







import unittest
import os
from pprint import pprint as pp

from mulgrids import *
from t2data import *
from t2listing import *

TESTDIR = './tests/data'

class TestObsDefs(unittest.TestCase):
    """ Test goPESTobs type by type, from users perspective. """
    def setUp(self):
        self.original_dir = os.getcwd()
        os.chdir(TESTDIR)

        self.geo = mulgrid("./gwai6307_06.dat")
        self.dat = t2data("./wai6307ns_021.dat")
        self.lst = t2listing("./wai6307ns_021.listing")

    def tearDown(self):
        del self.geo
        del self.dat
        self.lst.close()
        del self.lst
        os.chdir(self.original_dir)

    def test_totalheat_modelresult(self):
        """ totalheat should handle fixed/unfixed name even with regular expression """
        def fixname_and_re_problem(gnames):
            from gopest.obs import UserEntryObserv, OBS_USER_FUNC
            obsType = 'totalheat'
            fieldDataFile, customFilter, offsetTime = '', 'True', 0.0
            obsInfo = ["2000.00"] + gnames
            ue = UserEntryObserv(obsType,obsInfo,fieldDataFile,customFilter,
                offsetTime)

            obfs = OBS_USER_FUNC[ue.obsType+'_modelresult'](self.geo,self.dat,self.lst,ue)
            self.assertEqual(obfs[0], 42586.0, '%s failed to match "RH100"' % str(gnames))

        # all these SHOULD match exactly ONE gener: RH100
        fixname_and_re_problem(["'RH1 0'"])
        fixname_and_re_problem(["'RH(1 0|1 1)'"])
        fixname_and_re_problem(["'RH(100|101)'"])
        fixname_and_re_problem(["'RH1[ 01]0'"])
        fixname_and_re_problem(["'RH1[01]0'"])
        fixname_and_re_problem(["'RH1[ 1]0'"])
        fixname_and_re_problem(["'RH100'","'SO100'"]) # SO is MASS not HEAT

    def test_unique_name(self):
        """ function to generate unique obs names """
        from gopest.obs_def import unique_obs_name
        self.assertEqual("En_EE_45_0001", unique_obs_name("enthalpy", "EE 45"))
        self.assertEqual("En_EE_45_0002", unique_obs_name("enthalpy", "EE 45"))
        self.assertEqual("my_EE456_0001", unique_obs_name("myenthalpy", "EE[456]00"))

    def test_totalheat_raise_exp(self):
        """ totalheat should raise exception when creating obs, if specified geners does not match anything. """
        from gopest.obs import UserEntryObserv, OBS_USER_FUNC
        obsType = 'totalheat'
        fieldDataFile, customFilter, offsetTime = '', 'True', 0.0
        obsInfo = [
            "2000.00",
            "'XX100'",
            "'XX100'",
            "'XX100'",
            "'XX100'",
            "'XX100'",
            "'XX100'",
            ]
        # this should NOT match anything and raise exception
        ue = UserEntryObserv(obsType,obsInfo,fieldDataFile,customFilter,
            offsetTime)

        with self.assertRaises(Exception) as context:
            obss = OBS_USER_FUNC[ue.obsType+'_fielddata'](self.geo,self.dat,ue)
        self.assertTrue('does not match' in str(context.exception))


class FullTest_goPESTobs(unittest.TestCase):
    """ Test running goPESTobs.py as an app, and check basic function """
    def setUp(self):
        self.original_dir = os.getcwd()
        os.chdir(TESTDIR)

    def tearDown(self):
        os.chdir(self.original_dir)

    def generateInput(self, fname, lines):
        f = open(fname, 'w')
        f.write('\n'.join(lines))
        f.close()

    def linesFromOutput(self, fname, cleanup=False):
        """ cleanup=True will have all line ending removed and keeps only non-
        empty lines. """
        f = open(fname, 'r')
        lines = f.readlines()
        f.close()
        if cleanup:
            return [str(line.rstrip()) for line in lines if line.strip()]
        else:
            return lines

    def cleanFiles(self, filenames):
        import glob
        for f in filenames:
            for ff in glob.glob(f):
                os.remove(ff)

    def runFullTest(self, list_lines, ins_lines, obs_lines, obf_lines):
        """ given contents of goPESTobs.list file, goPESTobs.py will be run and
        results compared to expected ins/obs/obf.  Note all expected lines are
        cleaned-up and line endings removed. """
        self.generateInput('goPESTobs.list', list_lines)

        # to generate PEST .pst observation section and .ins:
        #      goPESTobs.py geo dat newPESTins
        # API:
        #      generate_obses_and_ins(fgeo, fdat, insToWrite,
        #                             fobses, fplts='goPESTobs.json', fcovs='goPESTobs.coverage')
        from gopest.obs import generate_obses_and_ins
        generate_obses_and_ins(
            "gwai6307_06.dat",
            "wai6307ns_021.dat",
            "pest_obs_ins",
            "pest_obs")

        ins = self.linesFromOutput("pest_obs_ins", cleanup=True)
        self.assertEqual(ins, ins_lines)
        obs = self.linesFromOutput("pest_obs", cleanup=True)
        self.assertEqual(obs, obs_lines)
        self.cleanFiles(["pest_obs_ins", "pest_obs"])

        # to read Tough2 results and write result file for PEST to read:
        #      goPESTobs.py geo dat lst newPESTobf
        # API:
        #      read_from_real_model(fgeo, fdat, flst, fobf, waiwera=False)
        from gopest.obs import read_from_real_model
        read_from_real_model(
            "gwai6307_06.dat",
            "wai6307ns_021.dat",
            "wai6307ns_021.listing",
            "pest_obs_obf")

        obf = self.linesFromOutput("pest_obs_obf", cleanup=True)
        self.assertEqual(obf, obf_lines)
        self.cleanFiles(["goPESTobs.list", "pest_obs_obf"])

        # TODO: review these additional, possibly some junk
        self.cleanFiles(["goPESTobs.coverage", "goPESTobs.json"])
        self.cleanFiles(["*.obs"])

    def test_totalheat(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "totalheat",
                "",
                "[Defaults]",
                "OBSNME = 'Heat'",
                "OBGNME = 'HtTotal'",
                "",
                "[Obs]",
                "2000.00",
                "'RH100'",
                "",
            ],
            [
                "pif #",
                "l1 [He_RH100_0001]21:41",
            ],
            [
                " He_RH100_0001         2.0000000000000e+03  1.00000e+00 HtTotal",
            ],
            [
                "He_RH100_0001         4.2586000000000e+04",
            ]
            )

    def test_totalupflow(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "totalupflow",
                "",
                "[Defaults]",
                "OBGNME = 'totalupflow'",
                "",
                "[Obs]",
                "1000.00",
                "'SO3 0'",
                "",
                "[Obs]",
                "2000.00",
                "'SO...'",
                "",
            ],
            [
                "pif #",
                "l1 [Uf_SO3_0_0001]21:41",
                "l1 [Uf_SO____0001]21:41",
            ],
            [
                " Uf_SO3_0_0001         1.0000000000000e+03  1.00000e+00 totalupflow",
                " Uf_SO____0001         2.0000000000000e+03  1.00000e+00 totalupflow",
            ],
            [
                "Uf_SO3_0_0001         1.0000000000000e+01",
                "Uf_SO____0001         7.0100000000000e+02",
            ]
            )

    def test_blocktemperature(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "blocktemperature",
                "",
                "[Defaults]",
                "OBSNME = 'TB'",
                "OBGNME = 'TempByBlock'",
                "",
                "[Obs]",
                "'surface'",
                "ex_obs_temp_by_block.dat",
                "",
                "# filter out unwanted blocks",
                "# also supports matching time",
                "[DataFilter]",
                "block not in ['BO993', 'BK900']",
                "[Obs]",
                "'surface', 100.0",
                "ex_obs_temp_by_block.dat",
                "",
            ],
            [
                "pif #",
                "l1 [TB_BO993_0001]21:41",
                "l1 [TB_BK900_0001]21:41",
                "l1 [TB_BJ_10_0001]21:41",
                "l1 [TB_BJ_10_0002]21:41",
            ],
            [
                " TB_BO993_0001         2.5000000000000e+01  1.00000e+00 TempByBlock",
                " TB_BK900_0001         1.0000000000000e+02  1.00000e+00 TempByBlock",
                " TB_BJ_10_0001         5.2000000000000e+01  1.00000e+00 TempByBlock",
                " TB_BJ_10_0002         5.2000000000000e+01  1.00000e+00 TempByBlock",
            ],
            [
                "TB_BO993_0001         1.5785000000000e+02",
                "TB_BK900_0001         4.3913000000000e+01",
                "TB_BJ_10_0001         8.5268000000000e+01",
                "TB_BJ_10_0002         8.5268000000000e+01",
            ]
            )

    def xtest_heatflow(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "heatflow",
                "",
                "[Defaults]",
                "OBSNME = 'Hf'",
                "OBGNME = 'hf'",
                "",
                "# each obs has two lines, first line is zone name, and second",
                "# line is expected value.",
                "[Obs]",
                "'AAA'",
                "100.0",
                "# The zone name should be a string in python style (in quotes),",
                "# which is name of user defined zone (defined in get_surface_heatflow.cfg,",
                "# each zone can be a polygon or a list of columns).  ",
                "",
                "[Obs]",
                "'BBB'",
                "200.0",
                "",
                "[Obs]",
                "'CCC'",
                "300.0",
                "",
            ],
            [
                "pif #",
                "l1 [Hf_AAA_0001]21:41",
                "l1 [Hf_BBB_0001]21:41",
                "l1 [Hf_CCC_0001]21:41",
            ],
            [
                " Hf_AAA_0001           1.0000000000000e+02  1.00000e+00 hf",
                " Hf_BBB_0001           2.0000000000000e+02  1.00000e+00 hf",
                " Hf_CCC_0001           3.0000000000000e+02  1.00000e+00 hf",
            ],
            [
                "Hf_AAA_0001           1.1348649000000e+02",
                "Hf_BBB_0001           8.9901714260000e+03",
                "Hf_CCC_0001           9.1036579160000e+03",
            ]
            )

    def xtest_heatflowminimum(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "heatflowminimum",
                "",
                "[Defaults]",
                "OBSNME = 'Hm'",
                "OBGNME = 'hfmin'",
                "",
                "# each obs has two lines, first line is zone name, and second",
                "# line is expected value.",
                "# The zone name should be a string in python style (in quotes),",
                "# which is name of user defined zone (defined in get_surface_heatflow.cfg,",
                "# each zone can be a polygon or a list of columns).  ",
                "[Obs]",
                "' 90'",
                "100.0",
                # model result is 1.1348649000000e+02
                "",
                "[Obs]",
                "'175'",
                "9000.0",
                # model result is 8.9901714260000e+03
                "",
            ],
            [
                "pif #",
                "l1 [Hm__90_0001]21:41",
                "l1 [Hm_175_0001]21:41",
            ],
            [
                " Hm__90_0001           1.0000000000000e+02  1.00000e+00 hfmin",
                " Hm_175_0001           9.0000000000000e+03  1.00000e+00 hfmin",
            ],
            [
                "Hm__90_0001           1.0000000000000e+02",
                "Hm_175_0001           8.9901714260000e+03",
            ]
            )

    def test_enthalpy(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "enthalpy",
                "",
                "[DataTimeOffset]",
                "1953.0 * 60.0 * 60.0 * 24.0 * 365.25",
                "",
                "[Defaults]",
                "OBSNME = 'En'",
                "OBGNME = 'enth14'",
                "",
                "[DataFilter]",
                "1970.0 <= time <= 1980.0",
                "",
                "[Obs]",
                "'SP 30'",
                "ex_obs_hist.dat",
                "",
                "[Defaults]",
                "OBGNME = 'enth45'",
                "",
                "[DataFilter]",
                "1972.0 <= time",
                "",
                "[Obs]",
                "'RW95[567]'",
                "ex_obs_hist.dat",
                "",
            ],
            [
                "pif #",
                "l1 [En_SP_30_0001]21:41",
                "l1 [En_SP_30_0002]21:41",
                "l1 [En_RW955_0001]21:41",
                "l1 [En_RW955_0002]21:41",
            ],
            [
                " En_SP_30_0001         1.0000000000000e+05  1.00000e+00 enth14",
                " En_SP_30_0002         2.0000000000000e+05  1.00000e+00 enth14",
                " En_RW955_0001         2.0000000000000e+05  1.00000e+00 enth45",
                " En_RW955_0002         3.0000000000000e+05  1.00000e+00 enth45",
            ],
            [
                "En_SP_30_0001         1.0401000000000e+06",
                "En_SP_30_0002         1.0401000000000e+06",
                "En_RW955_0001         8.3950000000000e+04",
                "En_RW955_0002         8.3950000000000e+04",
            ]
            )

    def test_pressure(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "pressure",
                "",
                "[DataTimeOffset]",
                "1953.0 * 60.0 * 60.0 * 24.0 * 365.25",
                "",
                "[Defaults]",
                "OBGNME = 'pressure'",
                "",
                "[DataFilter]",
                "1970.0 <= time <= 1980.0",
                "",
                "[Obs]",
                "'BC2 0'",
                "ex_obs_hist.dat",
                "",
                "[DataFilter]",
                "1972.0 <= time",
                "",
                "[Obs]",
                "'BC195'",
                "ex_obs_hist.dat",
                "",
            ],
            [
                "pif #",
                "l1 [Pr_BC2_0_0001]21:41",
                "l1 [Pr_BC2_0_0002]21:41",
                "l1 [Pr_BC195_0001]21:41",
                "l1 [Pr_BC195_0002]21:41",
            ],
            [
                " Pr_BC2_0_0001         1.0000000000000e+07  1.00000e+00 pressure",
                " Pr_BC2_0_0002         2.0000000000000e+07  1.00000e+00 pressure",
                " Pr_BC195_0001         2.0000000000000e+07  1.00000e+00 pressure",
                " Pr_BC195_0002         3.0000000000000e+07  1.00000e+00 pressure",
            ],
            [
                "Pr_BC2_0_0001         5.5193000000000e+06",
                "Pr_BC2_0_0002         5.5193000000000e+06",
                "Pr_BC195_0001         5.5419000000000e+06",
                "Pr_BC195_0002         5.5419000000000e+06",
            ]
            )

    def test_pressure_by_well(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "pressure_by_well",
                "",
                "[DataTimeOffset]",
                "1953.0 * 60.0 * 60.0 * 24.0 * 365.25",
                "",
                "[Defaults]",
                "OBGNME = 'pressure'",
                "",
                "[DataFilter]",
                "1970.0 <= time <= 1980.0",
                "",
                "[Obs]",
                "'TH  2', -125.0",
                "ex_obs_hist.dat",
                "",
                "[DataFilter]",
                "1972.0 <= time",
                "",
                "[Obs]",
                "'TH  3', -125.0",
                "ex_obs_hist.dat",
                "",
            ],
            [
                "pif #",
                "l1 [Pw_TH__2_0001]21:41",
                "l1 [Pw_TH__2_0002]21:41",
                "l1 [Pw_TH__3_0001]21:41",
                "l1 [Pw_TH__3_0002]21:41",
            ],
            [
                " Pw_TH__2_0001         1.0000000000000e+07  1.00000e+00 pressure",
                " Pw_TH__2_0002         2.0000000000000e+07  1.00000e+00 pressure",
                " Pw_TH__3_0001         2.0000000000000e+07  1.00000e+00 pressure",
                " Pw_TH__3_0002         3.0000000000000e+07  1.00000e+00 pressure",
            ],
            [
                "Pw_TH__2_0001         5.1317000000000e+06", # BC335
                "Pw_TH__2_0002         5.1317000000000e+06", # .
                "Pw_TH__3_0001         5.1267000000000e+06", # BC350
                "Pw_TH__3_0002         5.1267000000000e+06", # .
            ]
            )

    def test_temperature(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "temperature",
                "",
                "[DataTimeOffset]",
                "1953.0 * 60.0 * 60.0 * 24.0 * 365.25",
                "",
                "[Defaults]",
                "OBGNME = 'temperature'",
                "",
                "# Each [Obs] should have exactly TWO lines!",
                "# First line starts with a Well name as python (quoted) string,",
                "#   then followed by the time (number) after comma.",
                "# Second line is the name of the data file, also processed by user routines",
                "#   to become the actual data PEST compare against",
                "",
                "[DataFilter]",
                "elev < 350.0",
                "",
                "[Obs]",
                "'TM  1', 1967.00",
                "ex_obs_temp.dat",
                "",
                "# if time is not specified, 0.0 assumed, okay for NS results with single table ",
                "[Obs]",
                "'TM  2'",
                "ex_obs_temp.dat",
                "",
            ],
            [
                "pif #",
                "l1 [Tw_TM__1_0001]21:41",
                "l1 [Tw_TM__1_0002]21:41",
                "l1 [Tw_TM__1_0003]21:41",
                "l1 [Tw_TM__1_0004]21:41",
                "l1 [Tw_TM__2_0001]21:41",
                "l1 [Tw_TM__2_0002]21:41",
                "l1 [Tw_TM__2_0003]21:41",
                "l1 [Tw_TM__2_0004]21:41",
            ],
            [
                " Tw_TM__1_0001         2.0000000000000e+02  1.00000e+00 temperature",
                " Tw_TM__1_0002         2.0000000000000e+02  1.00000e+00 temperature",
                " Tw_TM__1_0003         2.0000000000000e+02  1.00000e+00 temperature",
                " Tw_TM__1_0004         2.0000000000000e+02  1.00000e+00 temperature",
                " Tw_TM__2_0001         2.0000000000000e+02  1.00000e+00 temperature",
                " Tw_TM__2_0002         2.0000000000000e+02  1.00000e+00 temperature",
                " Tw_TM__2_0003         2.0000000000000e+02  1.00000e+00 temperature",
                " Tw_TM__2_0004         2.0000000000000e+02  1.00000e+00 temperature",
            ],
            [
                "Tw_TM__1_0001         1.2131000000000e+02", # AP530
                "Tw_TM__1_0002         1.6224000000000e+02", # AR530
                "Tw_TM__1_0003         1.8296000000000e+02", # AT530
                "Tw_TM__1_0004         1.9329000000000e+02", # AV530
                "Tw_TM__2_0001         7.2344000000000e+01", # AP680
                "Tw_TM__2_0002         9.2872000000000e+01", # AR680
                "Tw_TM__2_0003         1.0999000000000e+02", # AT680
                "Tw_TM__2_0004         1.2302000000000e+02", # AV680
            ]
            )

    def test_temperature_thickness(self):
        self.runFullTest(
            [
                "[ObservationType]",
                "temperature_thickness",
                "",
                "[DataTimeOffset]",
                "1953.0 * 60.0 * 60.0 * 24.0 * 365.25",
                "",
                "[Defaults]",
                "OBGNME = 'temperature'",
                "",
                "# Each [Obs] should have exactly TWO lines!",
                "# First line starts with a Well name as python (quoted) string,",
                "#   then followed by the time (number) after comma.",
                "# Second line is the name of the data file, also processed by user routines",
                "#   to become the actual data PEST compare against",
                "",
                "[DataFilter]",
                "elev < 350.0",
                "",
                "[Obs]",
                "'TM  1', 1967.00",
                "ex_obs_temp.dat",
                "",
                "# if time is not specified, 0.0 assumed, okay for NS results with single table ",
                "[Obs]",
                "'TM  2'",
                "ex_obs_temp.dat",
                "",
            ],
            [
                "pif #",
                "l1 [Th_TM__1_0001]21:41",
                "l1 [Th_TM__1_0002]21:41",
                "l1 [Th_TM__1_0003]21:41",
                "l1 [Th_TM__1_0004]21:41",
                "l1 [Th_TM__2_0001]21:41",
                "l1 [Th_TM__2_0002]21:41",
                "l1 [Th_TM__2_0003]21:41",
                "l1 [Th_TM__2_0004]21:41",
            ],
            [
                " Th_TM__1_0001         2.0000000000000e+02  5.00000e+01 temperature",
                " Th_TM__1_0002         2.0000000000000e+02  5.00000e+01 temperature",
                " Th_TM__1_0003         2.0000000000000e+02  5.00000e+01 temperature",
                " Th_TM__1_0004         2.0000000000000e+02  5.00000e+01 temperature",
                " Th_TM__2_0001         2.0000000000000e+02  5.00000e+01 temperature",
                " Th_TM__2_0002         2.0000000000000e+02  5.00000e+01 temperature",
                " Th_TM__2_0003         2.0000000000000e+02  5.00000e+01 temperature",
                " Th_TM__2_0004         2.0000000000000e+02  5.00000e+01 temperature",
            ],
            [
                "Th_TM__1_0001         1.2131000000000e+02", # AP530
                "Th_TM__1_0002         1.6224000000000e+02", # AR530
                "Th_TM__1_0003         1.8296000000000e+02", # AT530
                "Th_TM__1_0004         1.9329000000000e+02", # AV530
                "Th_TM__2_0001         7.2344000000000e+01", # AP680
                "Th_TM__2_0002         9.2872000000000e+01", # AR680
                "Th_TM__2_0003         1.0999000000000e+02", # AT680
                "Th_TM__2_0004         1.2302000000000e+02", # AV680
            ]
            )


if __name__ == '__main__':
    """ test a single case by using case name as command argument, eg.:
            python goPESTtest.py FullTest_goPESTobs.test_temperature_thickness
    """
    unittest.main(verbosity=3)



#!/bin/python

import os
import subprocess

from gopest.common import config as cfg
from gopest.common import runtime

"""
Run this script to submit BeoPEST jobs on NeSI using Slurm.  This includes
master and slaves jobs.  Note the BeoPEST here uses TCP communication.

1. Modify the following paragraph of settings.
2. Set up PEST directory properly.
3. Run "python submit_beopest.py" from the main PEST directory.

Angus Yeh
a.yeh@auckland.ac.nz
June 2015
"""

PROJECT = cfg['nesi']['project']
PROJECT_MAUI = cfg['nesi']['maui']['project']
PROJECT_MAHUIKA = cfg['nesi']['mahuika']['project']

WALLTIME_MASTER = cfg['nesi']['walltime_master']
WALLTIME_SLAVES = WALLTIME_MASTER

WALLTIME_FORWARD = cfg['nesi']['walltime_forward']

MAUI_NTASKS = cfg['nesi']['maui']['ntasks']
MAHUIKA_NTASKS = cfg['nesi']['mahuika']['ntasks']

NUM_SLAVES = cfg['pest']['num_slaves']

MEM_PER_SLAVE = "5000" # MB
MEM_MASTER = "500"
MEM_FORWARD = "4000"

# SIMULATOR = "~/bin/autough2_6b" # specify absolute path if not in system path
# SIMULATOR = "waiwera-Mahuika" # specify absolute path if not in system path
# 'waiwera': local native waiwera, installed in path
# 'waiwera-dkr': running locally using Docker with pywaiwera installed
# 'waiwera-Maui': calling submit_beopest.py and use Maui
# 'waiwera-Mahuika': calling submit_beopest.py and use Mahuika
SIMULATOR = cfg['simulator']['executable']

PST_NAME = cfg['pest']['case-name']
PESTDIR = cfg['pest']['dir']
SLAVEDIR = cfg['pest']['slave_dirs']
BEOPEST = cfg['pest']['executable']
PORT = cfg['pest']['port']

SWITCHES = " ".join(cfg['pest']['switches'])
# additional swiches for beopest, eg /s for restart, /p1 for parallise 1st model run
# should not use /p1 with svda, no /p1 with obsreref_10 either
# /hpstart, make sure PST_NAME.hp exists
# /i jco reuse, make sure PST_NAME.jco exists
# NOTE it's possible to use /hpstart and /i together, will start update tests right away
# NOTE working directory is assumed to be where this script is launched

# if use /f, PEST_HP runs a sets of forward runs using .par files
F_PAR_SETS = ["prandom", "11", "50", "40", PST_NAME+".rrf"]
# Enter filename base of parameter value files:
# Enter first index to use:
# Enter last index to use:
# Enter parallel run packet size:
# Enter name for run results file:

# used for slurm to redirect as standard input
use_input = False
with open('_input', 'w') as f:
    # if PEST_HP asks name of jacobian file, then use file named PST_NAME + ".jco"
    if "/i" in SWITCHES:
        f.write(PST_NAME + ".jco\n")
        use_input = True
    if "/f" in SWITCHES:
        f.write("\n".join(F_PAR_SETS))
        use_input = True

SLAVES_ON_SCRATCH = False
KEEP_TARGZ_SLAVES = False
# normally when a slave ends, the slave directory will be tar-gz-ed for debugging
# but this can be huge due to TOUGH2 listing files, these files can be summarised
# (using head and tail) to save space
SUMMARY_LARGE_FILES = "" #"*.listing"

MAIN_DIR = os.getcwd()

# use absolute path if known
if PESTDIR:
    BEOPEST = os.path.join(PESTDIR,BEOPEST)

# list NeSI modules that needs to load for running scripts/AUTOUGH2 etc.
ENV_MODULES = [
    'module load gimkl/2018b',
    'module load Python-Geo/3.7.3-gimkl-2018b',
    'source /nesi/project/uoa00124/env-py3-gopest/bin/activate',
    # 18/12/2022 3:34:08 a.m.
    # # "module GCC/4.9.2",
    # # "module Python/2.7.11-foss-2015a",
    # "module load gimkl/2017a",
    # "module load GCC/7.1.0",
    # "module load Python-Geo/2.7.14-gimkl-2017a",
]
# user can load these modules by command 'source _load_modules.sh'
with open('_load_modules.sh', 'w') as f:
    f.write('\n'.join([
        "echo !!! Please source this file to modify environment in calling shell",
        "echo \"    'source _load_modules.sh'\"",
        ] + ENV_MODULES))

ENV_MODULES_MAUI = cfg['nesi']['maui']['env_init']
ENV_MODULES_MAHUIKA = cfg['nesi']['mahuika']['env_init']

# communication files:
# pest files
# goPEST*
# pest_model*
# real_model*       (from gopest.common.runtime)
# data_*            (from toml)
# gs_*              (from toml)


pst_name = cfg['pest']['case-name']
slave_files = [
    pst_name + '.pst',
    # communication files:
    "_input",
    "_pest_dir",
    "_tough2",
    "_logfile",
    "_master_out",
    # goPEST working files
    "pest_model.ins",
    "pest_model.tpl",
    "goPESTconfig.toml",
    "goPESTpar.list",
    "goPESTobs.list",
    # user supplied function, may not exist
    "goPESTuser.py",
]
# forward run files -> generated from toml
slave_files += [
    # runtime['filename']['save'],
    runtime['filename']['incon'],
    runtime['filename']['dat_orig'],
]
slave_files += runtime['filename']['all_geoms']
slave_files += runtime['filename']['dat_seq']
# slave_files += runtime['filename']['lst_seq']
# user files -> copied from toml
slave_files += cfg['files']['slave']

master_files = slave_files
if "/hpstart" in cfg['pest']['switches']:
    master_files.append(pst_name + '.hp')
if "/i" in cfg['pest']['switches']:
    master_files.append(pst_name + '.jco')

master_files += cfg['files']['master']

# All files needed in slave directory
# TODO: should have both master_files and slave_files
FILE_LIST = " ".join(slave_files)

"""
Details
-------

This explains the working solution of using BeoPEST on NeSI, which is
implemented by this script.

This version of BeoPEST works uses TCP, MPI not supported with some SVD stuff
John Doherty added (11/12/2014).

Slurm job files generated for submission:

    _master_job.sl   -> this is job file for launching master beopest along with
                        a set of slaves, using N+1 cpus
    _slaves_job.sl   -> job file for slaves, using N cpus
    _multi.conf      -> used by slaves job file, calls _run_a_slave.sh
    _run_master.sh   -> record which host it is running on, run master beopest
    _run_a_slave.sh  -> copy required files to unique slave directories, then
                        launch single beopest in slave mode.

    _master_dir      -> so slave jobs know where master directory is
    _master_out      -> so slave jobs know where to print information
    _pest_dir        -> so slaves know where to find PEST/BeoPEST utilities
    _tough2          -> so slaves know which SIMULATOR simulator to use

This set of scripts/job files basically launch a beopest as master in the main
directory by calling:

    beopest case /h :4004

In the master job file, a command is used to detect which host it is run on. The
info is saved to a file _master_host in the main direcotry.  Hence the slaves
are set to run with a small delay, to ensure the existence _master_host file.

Now a sbatch --dependency after:JOB_ID is used to launch slaves.  Note it's
after: not afterok: as seen in the examples.  This runs the slaves after master
STARTED (instead of finished).

The slaves job file uses srun --multi-prog, which reads _multi.conf file that
tells how to launch each slave.  Each slave is launched by using a bash script.
The script includes the code to:
    1. create unique slave dir, and copy required files
    2. change directory
    3. obtain the hostname of master beopest (from file _master_host)
    4. launch beopest case /h masterhost:4004

Also something bad might happen if more than one BeoPEST master uses the same
communication port.  So it's a good idea to set PORT to something different
when you receive this script.  Usually some large number is okay.

Tips For Users
--------------
    - You can launch more slave jobs by copy commands shown on screen after
      submission.  Note the dependency jobid is important.
    - Job ids are shown on the screen, you can cencel jobs by "scancel JOBID"
    - cleanups can be done by:
          rm -r slave_*
          rm *.out
          rm _*
      (be carefule not to delete your own files, if they match the names here.)
    - On linux, you have to change a script into an 'executable' if you want to
      run it directly like "./myscript.sh", instead of "bash myscript.sh".
      This requires you to issue the command:
          chmod +x myscript.sh
    - Within each of the slave directory, you can find a file called
      _master_dir, this file contains the master's working directory.  This is
      useful if some scripts running on a slave need to know where the master
      job is located.  (I use this to aid the transfer of updated .incon files
      across all slaves.)  You can use a command like this to get it back into
      a bash variable:
          MASTERDIR=`cat _master_dir`
          cp ${MASTERDIR}/a ./b
    - I use rsync to copy everything in a dir to another.  This is useful when
      you want to make some changes to a set of PEST setup/files.  Be careful
      with the slashes here.  eg.
          rsync -ar --progress case_1/ case_wai_2

NeSI related:
    - Read the first page when ssh into nesi login node.
    - use command show_my_projects to see what projects you are on, these
      proejcts directory should be where you keep model files etc.
    - compiling must be done on the build nodes, go to build node by ssh
      build-wm.
    - use private keys to simplify the login/scp procedure, on Windows I use
      Putty and related tools such as WinSCP etc.


TODOs
-----
    - make cleaning up easier

"""

# python 2.6 does not support the new/easier subprocess.check_output
def check_output_old(cmd):
    ''' so use os.system with "> /dev/null 2>&1" '''
    import tempfile
    handle, fname = tempfile.mkstemp(prefix='tmpout', dir='.', text=True)
    code = os.system(cmd + "> " + fname)
    with open(fname, 'r') as f:
        out = f.readlines()
    os.close(handle)
    os.remove(fname)
    if code != 0:
        print("Command %s failed: %s" % (cmd, "".join(out)))
    return "".join(out)

def check_output(cmd):
    return subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True).decode()

def write_to(filename, line):
    with open(filename, 'w') as f:
        f.write(line)

def get_master_dir():
    """ _master_dir is created and copied to slave dirs """
    try:
        with open('_master_dir', 'r') as f:
            line = f.readlines()[0].strip()
            return line
    except:
        return '..'

def get_slave_id():
    """ returns a string as id, needs to be called from within a 'slave' directory.
    NOTE _procid should be written by the master program that invokes slaves.
         Otherwise the current (sub) directory name will be used. """
    try:
        with open('_procid', 'r') as f:
            line = f.readlines()[0].strip()
            return line
    except:
        # get get only the last part of current working dir
        return os.path.basename(os.path.normpath(os.getcwd()))

def get_master_out():
    try:
        with open('_master_out', 'r') as f:
            line = f.readlines()[0].strip()
            return line
    except:
        return 'STDOUT'

def create_job_name():
    """ use directory's name, and make sure spaces replaced by _, for now. """
    name = os.path.split(check_output("pwd"))[-1].replace(' ','').replace('_','')
    if name.startswith('case'):
        name = 'c' + name[4:]
    return name.strip()

def gen_master_sl(fname="_master_job.sl"):
    jobname = create_job_name()
    txt = [
        "#!/bin/bash",
        "#SBATCH -J %s" % jobname,
        "#SBATCH -A %s         # Project Account" % PROJECT,
        "#SBATCH --time=%s     # Walltime" % WALLTIME_MASTER,
        # "#SBATCH --ntasks=1          # number of tasks",
        # "#SBATCH --mem=%s  # memory/cpu (in MB)" % MEM_MASTER,
        # # "#SBATCH --workdir=%s  # working dir" % MAIN_DIR,
        # "#SBATCH --ntasks=%i         # number of tasks" % int(NUM_SLAVES / 5),
        "#SBATCH --ntasks=1          # number of tasks",
        "#SBATCH --cpus-per-task=%i  # number of CPUs" % max(int(NUM_SLAVES / 5),1),
        "#SBATCH --overcommit        # allow many tasks on one CPU",
        "#SBATCH --mem=%i  # memory/cpu (in MB)" % (int(MEM_PER_SLAVE) * int(NUM_SLAVES / 5) + int(MEM_MASTER)),
        "#SBATCH --profile task",
        "#SBATCH --acctg-freq=1",
        ]
    if use_input:
        txt += ["#SBATCH --input=_input"]
    txt += [
        "",
        "echo running %s..." % fname,
        "",
        "rm -rf _jobs",
        "mkdir _jobs",
        "",
        "function finish {",
        "  cd _jobs",
        "  for f in *",
        "  do",
        "    echo EXIT $SLURM_JOB_ID, master script cancelling child job: $f",
        "    scancel --clusters=maui,mahuika $f",
        "  done",
        "  cd ..",
        "}",
        "trap finish EXIT",
        "",
        ]
    txt += ENV_MODULES
    txt += [
        "",
        "MASTERDIR=`pwd`",
        "echo ${MASTERDIR} > _master_dir",
        "",
        # "rm -f run_ns_pr.log",
        # "echo ${MASTERDIR}/run_ns_pr.log > _logfile",
        "",
        "# run master",
        # "srun --exclusive -n1 bash %s/_run_master.sh" % MAIN_DIR,
        "bash %s/_run_master.sh &" % MAIN_DIR,
        "",
        "",
        "# run slaves",
        "echo starting %i PEST slaves on single node..." % NUM_SLAVES,
        "for i in {1..%i}" % NUM_SLAVES,
        "do",
        "    bash _run_a_slave.sh ${i} &",
        "done",
        "",
        "# for both background shell jobs (master and slaves srun) to finish",
        "wait",
        "",
        ]
    with open(fname, 'w') as f:
        f.write("\n".join(txt))

def gen_slaves_sl(fname="_slaves_job.sl"):
    # be careful, might not work, if str contains txt such as 'MB'
    jobname = 'S_' + create_job_name()
    txt = "\n".join([
        "#!/bin/bash",
        "#SBATCH -J %s" % jobname,
        "#SBATCH -A %s         # Project Account" % PROJECT,
        "#SBATCH --time=%s     # Walltime" % WALLTIME_SLAVES,
        "#SBATCH --ntasks=1          # number of tasks",
        "#SBATCH --cpus-per-task=%i  # number of CPUs" % max(int(NUM_SLAVES / 10),1),
        "#SBATCH --overcommit        # allow many tasks on one CPU",
        "#SBATCH --mem=%i  # memory/cpu (in MB)" % (int(MEM_PER_SLAVE) * int(NUM_SLAVES / 5)),
        "#SBATCH --profile task",
        "#SBATCH --acctg-freq=1",
        # "#SBATCH --workdir=%s  # working dir" % MAIN_DIR,
        "echo running %s..." % fname,
        ] + ENV_MODULES + [
        "",
        "# run slaves",
        "echo starting %i PEST slaves on single node..." % NUM_SLAVES,
        "for i in {1..%i}" % NUM_SLAVES,
        "do",
        # "    echo slave id ${i}",
        "    bash _run_a_slave.sh ${i} &",
        "done",
        "",
        "# for both background shell jobs (master and slaves srun) to finish",
        "wait",
        "",
        ])
    f = open(fname, 'w')
    f.write(txt)
    f.close()

def gen_forward_sl(cmd, fname="_forward.sl"):
    cwd = os.getcwd()
    txt = "\n".join([
        "#!/bin/bash",
        "#SBATCH -J %s" % get_slave_id(),
        "#SBATCH -A %s    # Project Account" % PROJECT,
        "#SBATCH --time=%s    # Walltime" % WALLTIME_FORWARD,
        "#SBATCH --ntasks=1    # number of tasks",
        "#SBATCH --mem=%s    # memory/cpu (in MB)" % MEM_FORWARD,
        # "#SBATCH --workdir=%s    # working dir" % cwd,
        # "#SBATCH --output=%s # print to master .out" % get_master_out(),
        # "#SBATCH --error=%s # print to master .out" % get_master_out(),
        # "#SBATCH --open-mode=append # make sure don't overwrite master out",
        "",
        "function finish {",
        "  echo -- forward job $SLURM_JOB_ID exiting at $(pwd)",
        "  rm _status_on_nesi",
        "  rm %s/_jobs/$SLURM_JOB_ID" % get_master_dir(),
        "}",
        "trap finish EXIT",
        "",
        "",
        "echo -- forward job $SLURM_JOB_ID starting at $(pwd)",
        "touch %s/_jobs/$SLURM_JOB_ID" % get_master_dir(),
        "",
        ] + ENV_MODULES + [
        "",
        "srun %s" % cmd,
        "",
        ])
    with open(fname, 'w') as f:
        f.write(txt)

def gen_forward_mahuika_sl(cmd, fname="_forward.sl"):
    cwd = os.getcwd()
    txt = "\n".join([
        "#!/bin/bash -e",
        "#SBATCH -J %s" % get_slave_id(),
        "#SBATCH -A %s    # Project Account" % PROJECT_MAHUIKA,
        "#SBATCH --export=NONE      # don't carry env over",
        "#SBATCH --time=%s    # Walltime" % WALLTIME_FORWARD,
        "#SBATCH --ntasks=%i    # number of tasks" % MAHUIKA_NTASKS,
        "#SBATCH --mem=%s    # memory/cpu (in MB)" % MEM_FORWARD,
        #"#SBATCH --mem-per-cpu=%i  # memory/cpu (in MB)" % int(float(MEM_FORWARD)/float(MAHUIKA_NTASKS)),
        #"#SBATCH --mem=%s    # memory/cpu (in MB)" % MEM_FORWARD,
        #"#SBATCH --output=%s # print to master .out" % get_master_out(),
        #"#SBATCH --error=%s # print to master .out" % get_master_out(),
        #"#SBATCH --open-mode=append # make sure don't overwrite master out",
        "",
        "function finish {",
        "  echo -- forward job $SLURM_JOB_ID exiting at $(pwd)",
        "  rm _status_on_nesi",
        "  rm %s/_jobs/$SLURM_JOB_ID" % get_master_dir(),
        "}",
        "trap finish EXIT",
        "",
        "",
        "echo -- forward job $SLURM_JOB_ID starting at $(pwd)",
        "touch %s/_jobs/$SLURM_JOB_ID" % get_master_dir(),
        "",
        ] + ENV_MODULES_MAHUIKA + [
        "",
        "export SLURM_EXPORT_ENV=ALL",
        "srun %s" % cmd,
        "",
        ])
    with open(fname, 'w') as f:
        f.write(txt)


def gen_forward_maui_sl(cmd, fname="_forward.sl"):
    cwd = os.getcwd()
    txt = "\n".join([
        "#!/bin/bash -e",
        "#SBATCH -J %s" % get_slave_id(),
        "#SBATCH -A %s    # Project Account" % PROJECT_MAUI,
        "#SBATCH --clusters=maui    # from mahuika to maui",
        "#SBATCH --export=NONE      # don't carry env over",
        "#SBATCH --time=%s    # Walltime" % WALLTIME_FORWARD,
        "#SBATCH --ntasks=%i    # number of tasks" % MAUI_NTASKS,
        #"#SBATCH --mem-per-cpu=%i  # memory/cpu (in MB)" % int(float(MEM_FORWARD)/float(MAUI_NTASKS)),
        "#SBATCH --mem=%s    # memory/cpu (in MB)" % MEM_FORWARD,
        #"#SBATCH --output=%s # print to master .out" % get_master_out(),
        #"#SBATCH --error=%s # print to master .out" % get_master_out(),
        #"#SBATCH --open-mode=append # make sure don't overwrite master out",
        "#SBATCH --partition=nesi_research",
        #"#SBATCH --qos=nesi_debug",
        "",
        "function finish {",
        "  echo -- forward job $SLURM_JOB_ID exiting at $(pwd)",
        "  rm _status_on_nesi",
        "  rm %s/_jobs/$SLURM_JOB_ID" % get_master_dir(),
        "}",
        "trap finish EXIT",
        "",
        "",
        "echo -- forward job $SLURM_JOB_ID starting at $(pwd)",
        "touch %s/_jobs/$SLURM_JOB_ID" % get_master_dir(),
        "",
        ] + ENV_MODULES_MAUI + [
        "",
        "export SLURM_EXPORT_ENV=ALL",
        "srun %s" % cmd,
        "",
        ])
    with open(fname, 'w') as f:
        f.write(txt)

def gen_run_master(fname="_run_master.sh"):
    if use_input:
        cmd = "%s %s %s /h :%s < _input" % (BEOPEST, PST_NAME, SWITCHES, PORT)
    else:
        cmd = "%s %s %s /h :%s" % (BEOPEST, PST_NAME, SWITCHES, PORT)
    txt = "\n".join([
        "#!/bin/bash",
        "",
        "echo Master working at $(hostname)",
        "echo $(hostname) > _master_host",
        "",
        "echo $(pwd)/slurm-${SLURM_JOB_ID}.out > _master_out",
        "",
        "echo Master running command: %s" % cmd,
        cmd,
        "",
        ])
    f = open(fname, 'w')
    f.write(txt)
    f.close()

def gen_run_single_slave(fname="_run_a_slave.sh"):
    # use jobid and procid to ensure slave directories are unique.
    # add ./ into PATH so that PEST can run model files without ./
    # "slave_${SLURM_JOB_ID}_${SLURM_STEP_ID}_${SLURM_PROCID}_${SLURM_LOCALID}_$1"
    slave_name = "slave_${SLURM_JOB_ID}_$1"
    slave_dir = os.path.join(SLAVEDIR, slave_name)
    cmd = "%s %s /h ${MASTERHOST}:%s" % (BEOPEST, PST_NAME, PORT)
    lines = [
        "#!/bin/bash",
        "",
        "echo running %s $1 in %s..." % (fname, slave_dir),
        "",
        "sleep 5",
        "mkdir -pv %s" % slave_dir,
        # "cp * %s/" % slave_dir,
        "cp %s %s/" % (FILE_LIST, slave_dir),
        "cd %s" % slave_dir,
        "echo A slave working at $(hostname):%s" % slave_dir,
        "echo %s > _master_dir" % MAIN_DIR, # needed for model scripts to copy incon/save
        "echo ${SLURM_JOB_ID}_$1 > _procid", # needed for obsreref book keeping
        "MASTERHOST=`awk '{print $1}' %s/_master_host`" % MAIN_DIR,
        "export PATH=./:$PATH",
        "echo Slave $(hostname)_${1} running command: %s" % cmd,
        # "sysctl fs.file-nr",
        # "chmod a+x model.bat",
        # "chmod a+x r_model.bat",
        # "chmod a+x d_model.bat",
        cmd,
        "",]
    if SUMMARY_LARGE_FILES:
        lines += [
            "# head and tail big result files",
            "for f in %s" % SUMMARY_LARGE_FILES,
            "do",
            "  head --lines=20 $f > $f.head",
            "  tail --lines=20 $f > $f.tail",
            "  echo Removing $f...",
            "  rm $f",
            "done",
            "",]
    if KEEP_TARGZ_SLAVES:
        # lines += [
        #     "cd ..",
        #     "",]
        lines += [
            "tar -zcf %s/%s.tar.gz %s" % (MAIN_DIR, slave_name, slave_dir),
            "",]
    with open(fname, 'w') as f:
        f.write("\n".join(lines))

def gen_test_dir(fname="_copy_to_test.sh"):
    # make a test directory, used for manually checking slave directory
    slave_dir = "%s/test" % MAIN_DIR
    txt = "\n".join([
        "#!/bin/bash",
        "",
        "mkdir -pv %s" % slave_dir,
        "cp %s %s/" % (FILE_LIST, slave_dir),
        ])
    f = open(fname, 'w')
    f.write(txt)
    f.close()


def sbatch_check(cmd, retry_sec=None, retry_limit=10):
    """ this runs sbatch and detects errors, only return slurm job id if
    successful.  If retry is not None, sbatch will be called again after
    sleeping retry seconds, with a limit number of tries.
    """
    import sys
    if retry_sec is None:
        retry_limit = 0
    i = 0
    jobid = None
    while i < (retry_limit + 1):
        print("running %s ..." % cmd)
        out = check_output(cmd)
        i += 1
        try:
            jobid = out.strip().split()[3]
            break
        except IndexError:
            print("sbatch error, try no. %i: %s" % (i, out.strip()))
        except:
            print("sbatch unexpected error:", sys.exc_info()[0])
            raise
        #if 'error' in out:
        #    print("%i: %s" % (i, out.strip()))
        #else:
        #    jobid = out.strip().split()[3]
        #    break
    return jobid

def proc_args():
    # couldn't use argparse because NeSI Python still default to 2.6
    import sys
    def get_opt(o):
        for i,s in enumerate(sys.argv[1:]):
            if s == o:
                return sys.argv[1:][i+1]
    def usage():
        print('\n'.join([
            "This script submits BeoPEST/PEST_HP job on NeSI.",
            "  -f, --forward COMMAND",
            "      Submit single forward run, COMMAND must be supplied.",
            "      Optional, if not supplied the whole PEST master/slaves run",
            "      will be submitted instead",
            "  -f2, --forward2 COMMAND",
            "      This is the same as -f/--forward.  But instead using the",
            "      less reliable 'sbatch --wait', now it uses 'swait'.",
            "  -f3, --forward3 COMMAND",
            "      This is the same as -f/--forward.  But file locking",
            "      mechanism instead of swait/sbatch that replies on good",
            "      communication",
            "  -f3mahuika, --forward3mahuika COMMAND",
            "      This is the same as -f3/--forward3, but submit to Mahuika",
            "      instead.",
            "  -f3maui, --forward3maui COMMAND",
            "      This is the same as -f3/--forward3, but submit to Maui",
            "      instead.",
            "  -f3x, --forward3x COMMAND",
            "      This is the same as -f3/--forward3, but submission to",
            "      Maui or Mahuika depends on settings in goPESTconfig.toml",
            "  --dirs DIR_PATTERNi --jobnowait COMMAND",
            "      Used to submit many jobs of COMMAND in directories specified",
            "      by DIR_PATTERN",
            "  --cancel",
            "      Cancel ALL jobs originated from current directory (_jobs).",
            ]))
    option = {
        "forward": None,
        "forward2": None,
        "forward3": None,
        "forward3mahuika": None,
        "forward3maui": None,
        "forward3x": None,
        "dirs": None,
        "jobnowait": None,
        "cancel": False,
    }
    if '--help' in sys.argv[1:]:
        usage()
    elif '-h' in sys.argv[1:]:
        usage()
    elif '--forward' in sys.argv[1:]:
        option['forward'] = get_opt('--forward')
    elif '-f' in sys.argv[1:]:
        option['forward'] = get_opt('-f')
    elif '--forward2' in sys.argv[1:]:
        option['forward2'] = get_opt('--forward2')
    elif '-f2' in sys.argv[1:]:
        option['forward2'] = get_opt('-f2')
    elif '--forward3' in sys.argv[1:]:
        option['forward3'] = get_opt('--forward3')
    elif '-f3' in sys.argv[1:]:
        option['forward3'] = get_opt('-f3')
    elif '--forward3mahuika' in sys.argv[1:]:
        option['forward3mahuika'] = get_opt('--forward3mahuika')
    elif '-f3mahuika' in sys.argv[1:]:
        option['forward3mahuika'] = get_opt('-f3mahuika')

    elif '--forward3maui' in sys.argv[1:]:
        option['forward3maui'] = get_opt('--forward3maui')
    elif '-f3maui' in sys.argv[1:]:
        option['forward3maui'] = get_opt('-f3maui')
    elif '--forward3x' in sys.argv[1:]:
        option['forward3x'] = get_opt('--forward3x')
    elif '-f3x' in sys.argv[1:]:
        option['forward3x'] = get_opt('-f3x')
    elif '--dirs' in sys.argv[1:]:
        option['dirs'] = get_opt('--dirs')
        option['jobnowait'] = get_opt('--jobnowait')
    elif '--cancel' in sys.argv[1:]:
        option['cancel'] = True
    return option

def submit_cli(argv=[]):
    if cfg['mode'] != 'nesi':
        raise Exception('Error! gopest submit can only run with mode = "nesi" ')

    from time import sleep
    # You can use chain jobs to create dependencies between jobs.
    # SLURM has an option -d or "--dependency" that allows to
    # specify that a job is only allowed to start if another job finished.
    #
    # use 'after:' instead of 'afterok:' to start slaves after master STARTED
    # (not after it's done)

    SCRIPT_DIR = os.getcwd()
    option = proc_args()

    ### print basic info
    print("You are working under directory: %s" % SCRIPT_DIR)
    write_to('_pest_dir', PESTDIR)
    write_to('_tough2', SIMULATOR)
    write_to('_master_dir', SCRIPT_DIR)

    # for job management, keeps a list of jobs submitted from this directory
    if not os.path.exists('_jobs'):
        os.makedirs('_jobs')

    ### if forward only submit single run job
    if option['forward'] is not None:
        gen_forward_sl(option['forward'])
        # important to have --wait here to block until job finish
        jobid = sbatch_check("sbatch --wait _forward.sl", retry_sec=30)
        if jobid is not None:
            ttime = check_output("sacct -j %s.0 -o totalcpu -n" % jobid).strip()
            print("\nForward job %s finished after %s" % (jobid, ttime))
        else:
            print("\nFailed to submit _forward.sl")
        exit()
    elif option['forward2'] is not None:
        """ The is the same as -f/--forward, but use the more reliable swait.

        It happened a few times that sbatch --wait returned early with error:
        sbatch: error: slurm_receive_msg: Socket timed out on send/recv
        operation. This may be triggered when slurm daemon is acutely busy.  The
        actual job in this case can still be running.  So Gene (from NeSI)
        suggested using swait script.
        """
        gen_forward_sl(option['forward2'])
        jobid = check_output("sbatch _forward.sl").strip().split()[3]
        print("\nJob %s submitted." % jobid)
        print("/share/bin/swait %s" % jobid)
        os.system("/share/bin/swait %s" % jobid)
        ttime = check_output("sacct -j %s.0 -o totalcpu -n" % jobid).strip()
        print("\nForward job %s finished after %s" % (jobid, ttime))
        exit()
    elif option['forward3'] is not None:
        """ The is the same as -f/--forward, but use lock files instead of
        swait or sbatch --wait, which replies on NeSI's communications.
        """
        gen_forward_sl(option['forward3'])
        # _status_on_nesi will be removed once the _forward.sl finish
        # regardless how it terminated.
        with open('_status_on_nesi', 'w') as f:
            pass
        import random
        wait_t = random.random() # * 20.0 * 60.0
        print('.. waiting %f sec before submit ..' % wait_t)
        sleep(wait_t)
        jobid = sbatch_check("sbatch _forward.sl", retry_sec=30, retry_limit=50)
        if jobid is not None:
            while os.path.isfile('_status_on_nesi'):
                sleep(120)
            ttime = check_output("sacct -j %s.0 -o totalcpu -n" % jobid).strip()
            print("\nForward job %s finished after %s" % (jobid, ttime))
        else:
            os.remove('_status_on_nesi')
            print("\nFailed to submit _forward.sl")
        exit()
    elif option['forward3x'] is not None:
        """ The is the same as -f/--forward, but use lock files instead of
        swait or sbatch --wait, which replies on NeSI's communications.
        """
        if cfg['nesi']['cluster_forward'] == 'mahuika':
            cmd = cfg['nesi']['mahuika']['executable'] + ' ' + option['forward3x']
            print('submit_beopest.py runs command (mahuika)' + cmd)
            gen_forward_mahuika_sl(cmd)
        elif cfg['nesi']['cluster_forward'] == 'maui':
            cmd = cfg['nesi']['maui']['executable'] + ' ' + option['forward3x']
            print('submit_beopest.py runs command (maui)' + cmd)
            gen_forward_maui_sl(cmd)
        else:
            raise Exception('only supports mahuika or maui')
        # _status_on_nesi will be removed once the _forward.sl finish
        # regardless how it terminated.
        with open('_status_on_nesi', 'w') as f:
            pass
        import random
        wait_t = random.random() * 5.0 # * 20.0 * 60.0
        print('.. waiting %f sec before submit ..' % wait_t)
        sleep(wait_t)
        jobid = sbatch_check("sbatch _forward.sl", retry_sec=30, retry_limit=50)
        if jobid is not None:
            while os.path.isfile('_status_on_nesi'):
                sleep(30)
            ttime = check_output("sacct --clusters=mahuika -j %s.0 -o totalcpu -n" % jobid).strip()
            print("\nForward job %s finished after %s" % (jobid, ttime))
        else:
            os.remove('_status_on_nesi')
            print("\nFailed to submit _forward.sl")
        exit()
    elif option['forward3mahuika'] is not None:
        """ The is the same as -f/--forward, but use lock files instead of
        swait or sbatch --wait, which replies on NeSI's communications.
        """
        gen_forward_mahuika_sl(option['forward3mahuika'])
        # _status_on_nesi will be removed once the _forward.sl finish
        # regardless how it terminated.
        with open('_status_on_nesi', 'w') as f:
            pass
        import random
        wait_t = random.random() # * 20.0 * 60.0
        print('.. waiting %f sec before submit ..' % wait_t)
        sleep(wait_t)
        jobid = sbatch_check("sbatch _forward.sl", retry_sec=30, retry_limit=50)
        if jobid is not None:
            while os.path.isfile('_status_on_nesi'):
                sleep(120)
            ttime = check_output("sacct --clusters=mahuika -j %s.0 -o totalcpu -n" % jobid).strip()
            print("\nForward job %s finished after %s" % (jobid, ttime))
        else:
            os.remove('_status_on_nesi')
            print("\nFailed to submit _forward.sl")
        exit()
    elif option['forward3maui'] is not None:
        """ The is the same as -f/--forward, but use lock files instead of
        swait or sbatch --wait, which replies on NeSI's communications.
        """
        gen_forward_maui_sl(option['forward3maui'])
        # _status_on_nesi will be removed once the _forward.sl finish
        # regardless how it terminated.
        with open('_status_on_nesi', 'w') as f:
            pass
        import random
        wait_t = random.random() # * 20.0 * 60.0
        print('.. waiting %f sec before submit ..' % wait_t)
        sleep(wait_t)
        jobid = sbatch_check("sbatch _forward.sl", retry_sec=30, retry_limit=50)
        if jobid is not None:
            while os.path.isfile('_status_on_nesi'):
                sleep(120)
            ttime = check_output("sacct --clusters=maui -j %s.0 -o totalcpu -n" % jobid).strip()
            print("\nForward job %s finished after %s" % (jobid, ttime))
        else:
            os.remove('_status_on_nesi')
            print("\nFailed to submit _forward.sl")
        exit()
    elif option['dirs'] is not None and option['jobnowait'] is not None:
        import glob
        import shutil
        write_to('_submit.out', '\n'.join([
            'Running --dirs "%s" --jobnowait "%s"' % (option['dirs'], option['jobnowait']),
            '',
            ]))
        for d in glob.glob(option['dirs']):
            shutil.copy('_master_dir', d)
            shutil.copy('_tough2', d)
            cwd = os.getcwd()
            os.chdir(d)
            write_to('_master_out', cwd+'/_submit.out')
            gen_forward_sl(option['jobnowait'])
            import random
            wait_t = random.random() * 20.0 * 60.0
            print('.. waiting %f sec before submit ..' % wait_t)
            sleep(wait_t)
            jobid = sbatch_check("sbatch _forward.sl", retry_sec=30, retry_limit=50)
            os.chdir(cwd)
        exit()
    elif option['cancel'] is True:
        import glob
        for f in glob.glob('_jobs/*'):
            print('Cancelling %s' % os.path.basename(f))
            os.system('scancel --clusters=maui,mahuika %s' % os.path.basename(f))
        exit()
                

    ### generate slurm scripts etc.
    gen_master_sl()
    gen_slaves_sl()
    gen_run_master()
    gen_run_single_slave()
    gen_test_dir()

    # slave files and dir will be handled within _run_a_slave.sh

    ### submit beopest master job, and get job id, master.sl will record hostname
    out = check_output("sbatch _master_job.sl").strip()
    print("BeoPEST/PEST_HP Master and Slaves: ", out)
    dependency = out.split()[3]

    write_to('_master_slurm_id', dependency)

    ### submit beopest slaves job, depend on run after master started
    cmd = "sbatch --dependency after:%s _slaves_job.sl" % dependency
    #out = check_output(cmd).strip()
    #print("BeoPEST Slaves: ", out)
    print("Add more slaves (after master job is submitted) by using command:")
    print("    %s" % cmd)


if __name__ == "__main__":
    submit_cli()



import sys

from gopest import __version__

title = """
goPEST - Interfacing PEST with Waiwera and (AU)TOUGH2 simulators
"""

version = """Version: (%s)
""" % __version__

hlp = """
Usage: gopest COMMAND [ARGUMENTS]

Supported COMMANDs:
    help
    init [--no-copy][--no-par][--no-obs]    (make_case_pst)
    submit                                  (submit_beopest)
xx  run                                     (run_beopest)
    par                                     (goPESTpar)
    obs                                     (goPESTobs)
    run-pest-model                          (pest_model)
    run-forward                             (run_ns_pr)
    save-iter-files                         (rename_latest_files)
    check-slaves                            (check_slaves)

Important files for goPEST to work:
    goPESTconfig.toml
        This is the main configuration file.
    goPESTpar.list
        This controls model parameters included for PEST.
    goPESTobs.list
        This controls observations for PEST.

University of Auckland, 2012, 2022
"""

def gopest_cli():
    print(title)
    argc = len(sys.argv)
    if argc < 2:
        print(version + hlp)
    else:
        if sys.argv[1] == 'help':
            print(version + hlp)
        else:
            # NOTE loading gopest.common checks goPESTconfig.toml
            import gopest.common
            import gopest.par
            import gopest.obs
            import gopest.pest_model
            import gopest.run_ns_pr
            import gopest.submit_beopest
            import gopest.make_case_pst
            import gopest.rename_latest_files
            import gopest.check_slaves
            cmds = {
                'par': gopest.par.goPESTpar,
                'obs': gopest.obs.goPESTobs,
                'run-pest-model': gopest.pest_model.main_cli,
                'run-forward': gopest.run_ns_pr.main_cli,
                'submit': gopest.submit_beopest.submit_cli,
                'init': gopest.make_case_pst.make_case_cli,
                'save-iter-files': gopest.rename_latest_files.rename_latest_files,
                'check-slaves': gopest.check_slaves.check_slaves_cli,
            }
            if sys.argv[1] not in cmds:
                print(version + hlp)
                print('Error! COMMAND not recognised.')
                exit(1)
            cmds[sys.argv[1]](sys.argv[1:])

"""
- NOTE good reference on designing CLI command names:
  https://smallstep.com/blog/the-poetics-of-cli-command-names/

- TODO it's probably not necessary to have so many commands
"""



from multiprocessing import Pool, cpu_count
import time
import json
import os
import os.path
import sys
import shutil
import glob
import subprocess
import collections.abc

import yaml
import tomlkit
import xlwt

from gopest.common import config
from gopest.common import runtime

def nested_dict_update(d, u):
    """ update a nested dict object with an update dict
    https://stackoverflow.com/a/3233356/2368167
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = nested_dict_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d

def unix2dos(fname):
    import sys
    with open(fname, 'rb') as infile:
        instr = infile.read()
    outstr = instr.replace( b"\r\n", b"\n" ).replace( b"\r", b"\n" ).replace( b"\n", b"\r\n" )
    if len(outstr) == len(instr):
        return
    with open(fname, 'wb') as outfile:
        outfile.write( outstr )

def fixpcf_noptmax(fpst, noptmax=0):
    """ update PEST case file (.pst) with NOPTMAX, number of optimisation
    iterations
    """
    fpst_bk = fpst + '.backup'
    if not os.path.isfile(fpst):
        print('Error: %s does not exist.' % fpst)
        exit(1)
    if os.path.isfile(fpst_bk):
        os.remove(fpst_bk)
    os.rename(fpst, fpst_bk)

    try:
        with open(fpst_bk,'r') as fin, open(fpst, 'w') as fout:
            linesep = None # use current file's line ending
            isec = 7 # start with 7, so that it will never be 7 again
            for line in fin:
                # detect current file's line ending first
                if linesep is None:
                    if line[-2] == '\r' and line[-1] == '\n':
                        linesep = '\r\n'
                    elif line[-1] == '\n':
                        linesep = '\n'
                # find correct line, replace
                isec += 1
                oline = line
                if line.strip().startswith('*') and 'control data' in line:
                    isec = 0
                if isec == 7:
                    # replace first item in line (NOPTMAX)
                    sps = line.split()
                    oline = ' ' + ' '.join([str(noptmax)] + sps[1:] + [linesep])
                fout.write(oline)
    except Exception as e:
        fin.close()
        fout.close()
        print('fixpcf_noptmax() failed to proceed, restoring...')
        os.remove(fpst)
        os.rename(fpst_bk, fpst)
        raise(e)

def read_rec(rec_file):
    values = {}
    with open(rec_file, 'r') as f:
        reading = False
        for line in f:
            if reading:
                if empty_line:
                    empty_line = False
                    continue
                if not line.strip():
                    continue
                if '----->' in line:
                    break
                sps = line.strip().split('=')
                v = float(sps[-1])
                s = sps[0].strip()
                if s == 'Sum of squared weighted residuals (ie phi)':
                    k = 'phi'
                elif s == 'Optimised measurement objective function':
                    k = 'phi measured'
                elif s == 'Optimised regularisation objective function':
                    k =  'phi regularisation'
                elif '"' in s:
                    k = s.split('"')[1]
                else:
                    k, v = None, None
                if k is not None:
                    values[k] = v
            if line.startswith('Objective function ----->'):
                reading, empty_line = True, True
    return values

class message(object):
    """ A Waiwera log (.yaml) message object """
    def __init__(self, msg):
        self._msg = msg
        self.level, self.source, self.event, self.data = msg


def check_sim_ends(spath):
    """ extract summary of waiwera simulation from yaml log """
    def waiwera_ends(fyaml):
        """ return elapsed time in seconds of waiwera simulation time, None if not
        successful """
        try:
            with open(fyaml) as f:
                data = yaml.load(f, Loader=yaml.CSafeLoader)
                r = {''}
                if data[-3][:3] == ['info', 'timestep', 'end']:
                    time = data[-3][3]['time']
                    size = data[-3][3]['size']
                if data[-1][:3] == ['info', 'simulation', 'destroy']:
                    elapsed = data[-1][3]['elapsed_seconds']
                return {'endtime': time, 'stepsize': size, 'elapsed': elapsed}
        except:
            return {}
    cwd = os.getcwd()
    os.chdir(spath)
    print('Working in %s...' % spath)

    results = {'run': {}}
    sequence = config['model']['sequence']
    input_typ = config['simulator']['input-type']
    flsts = runtime['filename']['lst_seq']
    for i,seq in enumerate(sequence):
        results['run'][seq] = {}
        if input_typ== 'waiwera':
            fyaml = flsts[i].replace('.h5', '.yaml')
            results['run'][seq] = waiwera_ends(fyaml)
        else:
            raise NotImplementedError()

    os.chdir(cwd)
    # print('Restored directory %s' % cwd)
    return results

def check_run_status(spath):
    """ extract summary of waiwera simulation from yaml log """
    cwd = os.getcwd()
    os.chdir(spath)
    print('Working in %s...' % spath)

    results = {'run': {}}
    sequence = config['model']['sequence']
    input_typ = config['simulator']['input-type']
    flsts = runtime['filename']['lst_seq']
    for i,seq in enumerate(sequence):
        results['run'][seq] = {}
        results['run'][seq]['exists'] = os.path.exists(flsts[i])

    os.chdir(cwd)
    # print('Restored directory %s' % cwd)
    return results

def get_obj_fn(spath):
    """ Return PEST calculated objective function values by modifying config and
    PEST case file to run a dummy run.  Then read results from case .rec file.

    This is more destructive, but I have decided to leave those modification
    in place after running the command for a couple of reasons:
    1. (lazy) no need to handle undo the mod if exception is raised
    2. (safe) it's probably safer to leave as modified here, reduce possibility
       for user to wipe out existing model output files, which can be costly to
       run

    Unfortunately I need to modify the config.toml file, because gopest run-
    pest-model is involked by PEST, not this script directly.
    """
    cwd = os.getcwd()
    os.chdir(spath)
    print('Working in %s...' % spath)

    # only works if the final real model output file exists
    flst = runtime['filename']['lst_seq'][-1]
    if not os.path.exists(flst):
        print('Real model output file %s is missing, cannot get obj. fn.' % flst)
        os.chdir(cwd)
        return {}

    # modify config to skip model run
    with open('goPESTconfig.toml', 'r') as f:
        cfg = tomlkit.load(f)
    cfg['model']['skip'] = True
    with open('goPESTconfig.toml', 'w') as f:
        tomlkit.dump(cfg, f)

    # modify NOPTMAX in control data .pst file
    fpst = config['pest']['case-name'] + '.pst'
    fixpcf_noptmax(fpst, 0)

    cmd = [
        os.path.join(config['pest']['dir'], config['pest']['executable']),
        config['pest']['case-name'],
        ]
    print('Running: ', cmd)
    results = {}
    subprocess.call(cmd)
    results['obj-fn'] = read_rec('case_reg.rec')

    os.chdir(cwd)
    # print('Restored directory %s' % cwd)
    return results

def export_xls(data):
    phis, regs, others = [], [], []
    for sln,sl in data.items():
        if 'obj-fn' in sl:
            for n in sorted(sl['obj-fn'].keys()):
                if n.startswith('phi'):
                    phis.append(n)
                elif n.startswith('regul_'):
                    regs.append(n)
                else:
                    others.append(n)
            phis, regs, others = sorted(phis), sorted(regs), sorted(others)
            break
    runcs = []
    for sln,sl in data.items():
        if 'run' in sl:
            for rn,r in sl['run'].items():
                runcs += [rn+'.'+c for c in sorted(r.keys())]
            break
    cols = ['slave'] + phis + others + regs + runcs

    wb = xlwt.Workbook()
    ws = wb.add_sheet('slaves')

    i = 0
    for j,c in enumerate(cols):
        ws.write(i, j, c)

    for sln,sl in data.items():
        if 'obj-fn' in sl:
            i += 1
            j = 0
            ws.write(i, j, sln)
            for p in (phis + others + regs):
                j += 1
                if p in sl['obj-fn']:
                    ws.write(i, j, sl['obj-fn'][p])
            for r in runcs:
                j += 1
                rn,sn = r.split('.')
                if 'run' in sl and rn in sl['run'] and sn in sl['run'][rn]:
                    ws.write(i, j, sl['run'][rn][sn])
    wb.save('goPESTslaves.xls')

def init_slave(cfg, rt):
    """ used in multiprocessing Pool to initialise threads with the the updated
    gopest.common.config and gopest.common.runtime
    """
    print('Updating local thread common.config and common.runtime.')
    global config, runtime
    config = cfg
    runtime = rt

hlp = '''
Usage: gopest check-slaves [--help] [--status] [--end-time] [--obj-fn]
                           [--dir path_to_slaves] [--pest-exe pest_executable]
                           [--export-xls]

The check-slaves command searches through slave directories and obtain/collect
their running status etc.  By default, the pest.slave_dirs property from
goPESTconfig.toml is used. It is possible to specify a different directory by
using argument "--dir".

Extracted information will be dumped into a JSON file "goPESTslaves.json".  Note
that the command will only update/append results into the JSON file if exists.
It tries not to destroy whatever that is already in the file.

"--status" is the fastest option, which only checks if model output files exist
in the lsave directories.  (ie. did the runs produce any output files at all)

"--end-time" enables extraction of simulation end time (from output YAML file if
running waiwera as simulator).

"--obj-fn" is more destructive.  Within each slave directory, it will run PEST
(with the NOPTMAX set to 0) once with the internal settings set to skip actual
model runs.  This essentially runs PEST so that observations and objective
function will be extracted from model outputs within the slave directory.

"--pest-exe" is useful with the "--obj-fn" option when user have the slaves
directories in a different environment than where it was originally run.
pest_executable here can include the path to the executable if it's not already
in the system's PATH.

"--export-xls" is used to export contents of goPESTslaves.json into spreadsheet
file goPESTslaves.xls.  This command can be used alone as it will load existing
goPESTslaves.json.

This command runs locally, and does not utilise any slurm/srun/queue facilities.

This message can be printed with the "--help" argument.  It is possible to
include more than one option at a time.
'''

def check_slaves_cli(argv=[]):
    if '--help' in argv or len(argv) <= 1:
        print(hlp)
        exit(0)
    else:
        # overwrite slave_dirs
        if '--dir' in argv:
            iarg = argv.index('--dir') + 1
            try:
                spath = argv[iarg]
            except IndexError:
                raise Exception('--dir argument needs to be followed by path_to_slaves')
        else:
            spath = config['pest']['slave_dirs']
        # overwrite PEST exe
        if "--pest-exe" in argv:
            iarg = argv.index('--pest-exe') + 1
            try:
                pexe = argv[iarg]
            except IndexError:
                raise Exception('--pest-exe argument needs to be followed by pest_executable')
            # modify the shared module variable
            config['pest']['dir'] = ''
            config['pest']['executable'] = pexe
        else:
            # do nothing, keeps original setting
            pass
        # tasks
        nopts, tasks = 0, []
        if "--status" in argv:
            nopts += 1
            tasks.append('status')
        if "--end-time" in argv:
            nopts += 1
            tasks.append('end-time')
        if "--obj-fn" in argv:
            nopts += 1
            tasks.append('obj-fn')
        xls = False
        if "--export-xls" in argv:
            nopts += 1
            xls = True
        if nopts == 0:
            print(hlp)
            print('Please specify at least one task to perform.')
            exit(0)

    start_time = time.time()
    if not all([os.path.exists(spath), os.path.isdir(spath)]):
        raise Exception('Specified path_to_slaves needs to be a valid directory: %s' % spath)

    task_fn = {
        'end-time': check_sim_ends,
        'status': check_run_status,
        'obj-fn': get_obj_fn,
    }

    slave_paths = sorted([p for p in glob.glob(os.path.join(spath, '*')) if os.path.isdir(p)])
    slave_names = [os.path.basename(sp) for sp in slave_paths]

    ncpu = max(1, cpu_count() - 1)
    print('Starting %i workers to process %i slaves' % (ncpu, len(slave_paths)))
    pool = Pool(ncpu, initializer=init_slave, initargs=(config, runtime))

    fout = 'goPESTslaves.json'
    if os.path.exists(fout):
        with open(fout, 'r') as f:
            data = json.load(f)
    else:
        data = {}

    for task in tasks:
        print('Checking %s...' % task)
        results = pool.map(task_fn[task], slave_paths)
        data = nested_dict_update(data, dict(zip(slave_names, results)))

    with open(fout, 'w') as f:
        json.dump(data, f, indent=4, sort_keys=True)

    if xls:
        export_xls(data)

    print('Finished after %f seconds' % (time.time() - start_time))

if __name__ == '__main__':
    check_slaves_cli(sys.argv)
